"""
.. module:: controller
   :synopsis: Basic controller classes for asynchronous optimization.
.. moduleauthor:: David Bindel <bindel@cornell.edu>

Modified to fit GridCal's purposes when using PySOT
"""

try:
    import Queue
except ImportError:
    import queue as Queue

import sys
import heapq
import threading
import logging
import time
from GridCal.Engine.Replacements.strategy import EvalRecord


# Get module-level logger
logger = logging.getLogger(__name__)


class Controller(object):
    """Base class for controller.

    Attributes:
        strategy: Strategy for choosing optimization actions.
        fevals: Database of function evaluations.
        feval_callbacks: List of callbacks to execute on new eval record
        term_callbacks: List of callbacks to execute on termination
    """

    def __init__(self):
        "Initialize the controller."
        logger.debug("Initialize controller")
        self.strategy = None
        self.fevals = []
        self.feval_callbacks = []
        self.term_callbacks = []

    def add_timer(self, timeout, callback):
        """Add a task to be executed after a timeout (e.g. for monitoring).

        Args:
            timeout: Time to wait before execution
            callback: Function to call when timeout elapses
        """
        thread = threading.Timer(timeout, callback)
        thread.daemon = True
        thread.start()

    def ping(self):
        "Tell controller to consult strategies when possible (if asynchronous)"
        pass

    def can_work(self):
        "Return whether we can currently perform work."
        return True

    def best_point(self, merit=None, filter=None):
        """Return the best point in the database satisfying some criterion.

        Args:
            merit: Function to minimize (default is r.value)
            filter: Predicate to use for filtering candidates

        Returns:
            Record minimizing merit() and satisfying filter();
            or None if nothing satisfies the filter
        """
        if filter is None:
            fcomplete = [f for f in self.fevals if f.is_completed]
        else:
            fcomplete = [f for f in self.fevals
                         if f.is_completed and filter(f)]
        if merit is None:
            def merit(r):
                return r.value
        if fcomplete:
            return min(fcomplete, key=merit)

    def new_feval(self, params, extra_args=None):
        """Add a function evaluation record to the database.

        In addition to adding the record with status 'pending',
        we run the feval_callbacks on the new record.

        Args:
            params: Parameters to the objective function

        Returns:
            New EvalRecord object
        """
        record = EvalRecord(params, extra_args=extra_args, status='pending')
        self.fevals.append(record)
        logger.debug("Call new feval callbacks")
        for callback in self.feval_callbacks:
            callback(record)
        return record

    def call_term_callbacks(self):
        "Call termination callbacks."
        logger.debug("Call termination callbacks")
        for callback in self.term_callbacks:
            callback()

    def add_term_callback(self, callback):
        "Add a callback for cleanup on termination."
        self.term_callbacks.append(callback)

    def add_feval_callback(self, callback):
        "Add a callback for notification on new fevals."
        self.feval_callbacks.append(callback)

    def remove_term_callback(self, callback):
        "Remove a callback from the term callback list."
        self.term_callbacks = [
            c for c in self.term_callbacks if c != callback
        ]

    def remove_feval_callback(self, callback):
        "Remove a callback from the feval callback list."
        self.feval_callbacks = [
            c for c in self.feval_callbacks if c != callback
        ]


class SerialController(Controller):
    """Serial optimization controller.

    Attributes:
        strategy: Strategy for choosing optimization actions.
        objective: Objective function
        fevals: Database of function evaluations
        skip: if True, skip over "None" proposals
    """

    def __init__(self, objective, skip=False):
        """Initialize the controller.

        Args:
            objective: Objective function
            skip: if True, skip over "None" proposals
        """
        Controller.__init__(self)
        self.objective = objective
        self.skip = skip

    def _run(self, merit=None, filter=None, reraise=True, stop_at=False, stop_value=0):
        "Run the optimization and return the best value."
        while True:
            proposal = self.strategy.propose_action()
            if not proposal:
                if not self.skip:
                    raise NameError('No proposed action')
            elif proposal.action == 'terminate':
                logger.debug("Accept termination proposal")
                proposal.accept()
                return self.best_point(merit=merit, filter=filter)
            elif proposal.action == 'eval':
                logger.debug("Accept eval proposal")
                proposal.record = self.new_feval(proposal.args)
                proposal.accept()
                try:
                    value = self.objective(*proposal.record.params)
                    proposal.record.complete(value)

                    # brute exit
                    if stop_at:
                        if value == stop_value:
                            return self.best_point(merit=merit, filter=filter)

                except Exception:
                    logger.exception("Error calling objective", exc_info=sys.exc_info())
                    proposal.record.cancel()
                    if reraise:
                        raise
            else:
                logger.debug("Reject proposal")
                proposal.reject()

    def run(self, merit=None, filter=None, reraise=True, stop_at=False, stop_value=0):
        """Run the optimization and return the best value.

        Args:
            merit: Function to minimize (default is r.value)
            filter: Predicate to use for filtering candidates
            reraise: Flag indicating whether exceptions in the
              objective function evaluations should be re-raised,
              terminating the optimization.

        Returns:
            Record minimizing merit() and satisfying filter();
            or None if nothing satisfies the filter
        """
        try:
            return self._run(merit=merit, filter=filter, reraise=reraise, stop_at=stop_at, stop_value=stop_value)
        finally:
            self.call_term_callbacks()


class ThreadController(Controller):
    """Thread-based optimization controller.

    The optimizer dispatches work to a queue of workers.
    Each worker has methods of the form

       worker.eval(record)
       worker.kill(record)

    These methods are asynchronous: they start a function evaluation
    or termination, but do not necessarily complete it.  The worker
    must respond to eval requests, but may ignore kill requests.  On
    eval requests, the worker should either attempt the evaluation or
    mark the record as killed.  The worker sends status updates back
    to the controller in terms of lambdas (executed at the controller)
    that update the relevant record.  When the worker becomes
    available again, it should use add_worker to add itself back to
    the queue.

    Attributes:
        strategy: Strategy for choosing optimization actions.
        fevals: Database of function evaluations
        workers: Queue of available worker threads
        messages: Queue of messages from workers

    """

    def __init__(self):
        "Initialize the controller."
        Controller.__init__(self)
        self.workers = Queue.Queue()
        self.messages = Queue.Queue()

    def ping(self):
        "Tell controller to consult strategies when possible"
        logger.debug("Wake thread controller")
        self.add_message()

    def add_timer(self, timeout, callback):
        """Add a task to be executed after a timeout (e.g. for monitoring).

        Args:
            timeout: Time to wait before execution
            callback: Function to call when timeout elapses
        """
        thread = threading.Timer(timeout, lambda: self.add_message(callback))
        thread.start()

    def add_message(self, message=None):
        """Queue up a message.

        Args:
            message: callback function with no arguments or None (default)
                     if None, a dummy message is queued to ping the controller
        """
        if message is None:
            self.messages.put(lambda: None)
        else:
            self.messages.put(message)

    def add_worker(self, worker):
        """Add a worker and queue a 'wake-up' message.

        Args:
            worker: a worker thread object
        """
        logger.debug("Add worker to thread controller")
        self.workers.put(worker)
        self.ping()

    def launch_worker(self, worker, daemon=False):
        """Launch and take ownership of a new worker thread.

        Args:
            worker: a worker thread object
            daemon: if True, the worker is launched in a daemon thread
                    (default is False)
        """
        logger.debug("Launch worker in thread controller")
        self.add_worker(worker)
        self.add_term_callback(worker.terminate)
        worker.daemon = worker.daemon or daemon
        worker.start()

    def can_work(self):
        "Claim we can work if a worker is available."
        return not self.workers.empty()

    def _submit_work(self, proposal):
        "Submit proposed work."
        try:
            worker = self.workers.get_nowait()
            logger.debug("Accept eval proposal")
            proposal.record = self.new_feval(proposal.args)
            proposal.record.worker = worker
            proposal.accept()
            worker.eval(proposal.record)
        except Queue.Empty:
            logger.debug("Reject eval proposal -- no worker")
            proposal.reject()

    def _run_message(self):
        "Process a message, blocking for one if none is available."
        message = self.messages.get()
        message()

    def _run_queued_messages(self):
        "Process any queued messages."
        while not self.messages.empty():
            self._run_message()

    def _run(self, merit=None, filter=filter):
        "Run the optimization and return the best value."
        while True:
            self._run_queued_messages()
            time.sleep(0)  # Yields to other threads
            proposal = self.strategy.propose_action()
            if not proposal:
                self._run_message()
            elif proposal.action == 'terminate':
                logger.debug("Accept terminate proposal")
                proposal.accept()
                return self.best_point(merit=merit, filter=filter)
            elif proposal.action == 'eval' and self.can_work():
                self._submit_work(proposal)
            elif proposal.action == 'kill' and not proposal.args[0].is_done:
                logger.debug("Accept kill proposal")
                record = proposal.args[0]
                proposal.accept()
                record.worker.kill(record)
            else:
                logger.debug("Reject proposal")
                proposal.reject()

    def run(self, merit=None, filter=None):
        """Run the optimization and return the best value.

        Args:
            merit: Function to minimize (default is r.value)
            filter: Predicate to use for filtering candidates

        Returns:
            Record minimizing merit() and satisfying filter();
            or None if nothing satisfies the filter
        """
        try:
            return self._run(merit=merit, filter=filter)
        finally:
            self.call_term_callbacks()


class BaseWorkerThread(threading.Thread):
    """Worker base class for use with the thread controller.

    The BaseWorkerThread class has a run routine that actually handles
    the worker event loop, and a set of helper routines for
    dispatching messages into the worker event loop (usually from the
    controller) and dispatching messages to the controller (usually
    from the worker).
    """

    def __init__(self, controller):
        "Initialize the worker."
        logger.debug("Initialize worker thread")
        super(BaseWorkerThread, self).__init__()
        self.controller = controller
        self.queue = Queue.Queue()

    def eval(self, record):
        """Start evaluation.

        Args:
            record: Function evaluation record
        """
        logger.debug("Queue eval at worker")
        self.queue.put(('eval', record))

    def kill(self, record):
        """Send kill message to worker.

        Args:
            record: Function evaluation record
        """
        logger.debug("Queue kill at worker")
        self.queue.put(('kill', record))

    def terminate(self):
        """Send termination message to worker.

        NB: if the worker is not running in a daemon thread,
        a call to the terminate method only returns after the
        the thread has terminated.
        """
        logger.debug("Queue terminate at worker")
        self.queue.put(('terminate',))
        if self.daemon:
            logger.debug("Do not wait on thread join (daemon)")
        else:
            logger.debug("Wait on worker thread join")
            self.join()

    def add_message(self, message):
        "Send message to be executed at the controller."
        self.controller.add_message(message)

    def add_worker(self):
        "Add worker back to the work queue."
        logger.debug("Worker thread is ready")
        self.controller.add_worker(self)

    def finish_success(self, record, value):
        "Finish successful work on a record and add ourselves back."
        logger.debug("Finished feval successfully")
        self.add_message(lambda: record.complete(value))
        self.add_worker()

    def finish_killed(self, record):
        "Finish recording killed on a record and add ourselves back."
        logger.debug("Feval killed")
        self.add_message(record.kill)
        self.add_worker()

    def finish_cancelled(self, record):
        "Finish recording cancelled on a record and add ourselves back."
        logger.debug("Feval cancelled")
        self.add_message(record.cancel)
        self.add_worker()

    def handle_eval(self, record):
        "Process an eval request."
        pass

    def handle_kill(self, record):
        "Process a kill request"
        pass

    def handle_terminate(self):
        "Handle any cleanup on a terminate request"
        pass

    def run(self):
        "Run requests as long as we get them."
        while True:
            request = self.queue.get()
            if request[0] == 'eval':
                logger.debug("Worker thread received eval request")
                record = request[1]
                self.add_message(record.running)
                self.handle_eval(record)
            elif request[0] == 'kill':
                logger.debug("Worker thread received kill request")
                self.handle_kill(request[1])
            elif request[0] == 'terminate':
                logger.debug("Worker thread received terminate request")
                self.handle_terminate()
                logger.debug("Exit worker thread run()")
                return


class BasicWorkerThread(BaseWorkerThread):
    """Basic worker for use with the thread controller.

    The BasicWorkerThread calls a Python objective function
    when asked to do an evaluation.  This is concurrent, but only
    results in parallelism if the objective function implementation
    itself allows parallelism (e.g. because it communicates with
    an external entity via a pipe, socket, or whatever).
    """

    def __init__(self, controller, objective):
        "Initialize the worker."
        super(BasicWorkerThread, self).__init__(controller)
        self.objective = objective

    def handle_eval(self, record):
        try:
            value = self.objective(*record.params)
            self.finish_success(record, value)
            logger.debug("Worker finished feval successfully")
        except Exception:
            self.finish_cancelled(record)
            logger.debug("Worker feval exited with exception")


class ProcessWorkerThread(BaseWorkerThread):
    """Subprocess worker for use with the thread controller.

    The ProcessWorkerThread is meant for use as a base class.
    Implementations that inherit from ProcessWorkerThread should
    define a handle_eval method that sets the process field so that it
    can be interrupted if needed.  This allows use of blocking
    communication primitives while at the same time allowing
    interruption.
    """

    def __init__(self, controller):
        "Initialize the worker."
        super(ProcessWorkerThread, self).__init__(controller)
        self.process = None

    def _kill_process(self):
        if self.process is not None and self.process.poll() is None:
            logger.debug("ProcessWorker is killing subprocess")
            self.process.terminate()

    def kill(self, record):
        "Send kill message."
        self._kill_process()
        super(ProcessWorkerThread, self).kill(record)

    def terminate(self):
        "Send termination message."
        self._kill_process()
        super(ProcessWorkerThread, self).terminate()


class SimTeamController(Controller):
    """Simulated parallel optimization controller.

    Run events in simulated time.  If two events are scheduled at the
    same time, we prioritize by when the event was added to the queue.

    Attributes:
        strategy: Strategy for choosing optimization actions.
        objective: Objective function
        delay: Time delay function
        workers: Number of workers available
        fevals: Database of function evaluations
        time: Current simulated time
        time_events: Time-stamped event heap
    """

    def __init__(self, objective, delay, workers):
        """Initialize the controller.

        Args:
            objective: Objective function
            delay: Time delay function (takes no arguments)
            workers: Number of workers available in simulation
        """
        Controller.__init__(self)
        self.objective = objective
        self.delay = delay
        self.workers = workers
        self.time = 0
        self.time_events = []
        self.event_id = 0

    def can_work(self):
        "Check if there are workers available."
        return self.workers > 0

    def submit_work(self, proposal):
        "Submit a work event."
        logger.debug("Accept eval proposal")
        self.workers -= 1
        record = self.new_feval(proposal.args)
        proposal.record = record
        proposal.accept()

        def event():
            "Closure for marking record done at some later point."
            if not record.is_done:
                try:
                    record.complete(self.objective(*record.params))
                    logger.debug("Finished evaluation successfully")
                except Exception:
                    record.cancel()
                    logger.debug("Finished evaluation with exception")
                self.workers += 1

        self.add_timer(self.delay(record), event)

    def kill_work(self, proposal):
        "Submit a kill event."
        logger.debug("Accept kill proposal")
        record = proposal.args[0]
        proposal.accept()

        def event():
            """Closure for canceling a function evaluation
            NB: This is a separate event because it will eventually have delay!
            """
            if not record.is_done:
                logger.debug("Finished killing evaluation")
                record.kill()
                self.workers += 1

        self.add_timer(0, event)

    def advance_time(self):
        "Advance time to the next event."
        assert self.time_events, "Deadlock detected!"
        event_time, event_id, event = heapq.heappop(self.time_events)
        self.time = event_time
        event()

    def add_timer(self, timeout, event):
        """Add a task to be executed after a timeout (e.g. for monitoring).

        Args:
            timeout: Time to wait before execution
            callback: Function to call when timeout elapses
        """
        heapq.heappush(self.time_events,
                       (self.time + timeout, self.event_id, event))
        self.event_id += 1

    def _run(self, merit=None, filter=None):
        "Run the optimization and return the best value."
        while True:
            proposal = self.strategy.propose_action()
            if not proposal:
                logger.debug("Advance")
                self.advance_time()
            elif proposal.action == 'terminate':
                logger.debug("Accepted terminate proposal")
                proposal.accept()
                return self.best_point(merit=merit, filter=filter)
            elif proposal.action == 'eval' and self.can_work():
                self.submit_work(proposal)
            elif proposal.action == 'kill' and not proposal.args[0].is_done:
                self.kill_work(proposal)
            else:
                logger.debug("Reject proposal")
                proposal.reject()
                self.advance_time()

    def run(self, merit=None, filter=None):
        """Run the optimization and return the best value.

        Args:
            merit: Function to minimize (default is r.value)
            filter: Predicate to use for filtering candidates

        Returns:
            Record minimizing merit() and satisfying filter();
            or None if nothing satisfies the filter
        """
        try:
            return self._run(merit=merit, filter=filter)
        finally:
            self.call_term_callbacks()


class ScriptedController(Controller):
    """Run a test script of actions from the controller.

    The ScriptedController is meant to test that a strategy adheres
    to an expected sequence of proposed actions in a given scenario.

    Attributes:
        strategy: Strategy for choosing optimization actions.
        fevals: Database of function evaluations
    """

    def __init__(self):
        Controller.__init__(self)
        self._can_work = True

    def add_timer(self, timeout, callback):
        "Add timer."
        assert False, "Timers not available in ScriptedController."

    def can_work(self):
        "Return True if worker available."
        return self._can_work

    def proposal(self, skip=False):
        """Return strategy proposal.

        Args:
            skip: if True, skip over all None proposals
        """
        proposal = self.strategy.propose_action()
        while skip and proposal is None:
            proposal = self.strategy.propose_action()
        return proposal

    def set_worker(self, v):
        "Set worker availability status."
        self._can_work = v

    def check_eval(self, proposal, args=None, pred=None):
        """Check whether a proposal is an expected eval proposal.

        Args:
            proposal: proposal to check
            args: expected evaluation args (if not None)
            pred: test predicate to run on proposal (if not None)
        """
        assert proposal is not None, \
            "Expected eval, got None"
        assert proposal.action == 'eval', \
            "Expected eval, got {0}".format(proposal.action)
        if args is not None:
            assert proposal.args == args, \
                "Expected eval at {0}, got {1}".format(args, proposal.args)
        if pred is not None:
            assert pred(proposal), \
                "Eval at {0} does not fit predicate".format(args)
        return proposal

    def check_kill(self, proposal, r=None):
        """Check whether a proposal is an expected kill proposal.

        Args:
            proposal: proposal to check
            r: record to be killed (or None if no check)
        """
        assert proposal is not None, \
            "Expected eval, got None"
        assert proposal.action == 'kill', \
            "Expected kill, got {0}".format(proposal.action)
        if r is not None:
            assert proposal.args[0] == r, \
                "Expected kill, but not at {0}".format(r.params)
        return proposal

    def check_terminate(self, proposal):
        """Check whether a proposal is an expected terminate proposal.

        Args:
            proposal: proposal to check
        """
        assert proposal is not None, \
            "Expected terminate, got None"
        assert proposal.action == 'terminate', \
            "Expected terminate, got {0}".format(proposal.action)
        return proposal

    def no_proposal(self):
        "Assert that next proposed action is None."
        logger.debug("Script: No proposal")
        assert self.proposal() is None

    def accept_eval(self, args=None, pred=None, skip=False):
        """Assert next proposal is an eval, which we accept.

        Args:
            args: expected evaluation args (if not None)
            pred: test predicate to run on proposal (if not None)
            skip: if True, skip over all None proposals

        Returns:
            proposal record
        """
        logger.debug("Script: accept proposal")
        proposal = self.proposal(skip=skip)
        proposal = self.check_eval(proposal, args=args, pred=pred)
        proposal.record = self.new_feval(proposal.args)
        proposal.accept()
        return proposal.record

    def accept_kill(self, r=None, skip=False):
        """Assert next proposal is a kill, which we accept.

        Args:
            r: record to be killed.
            skip: if True, skip over all None proposals
        """
        logger.debug("Script: accept kill")
        self.check_kill(self.proposal(skip=skip), r).accept()

    def accept_terminate(self, skip=False):
        """Assert next proposal is a kill, which we accept.

        Args:
            skip: if True, skip over all None proposals
        """
        logger.debug("Script: accept terminate")
        self.check_terminate(self.proposal(skip=skip)).accept()

    def reject_eval(self, args=None, pred=None, skip=False):
        """Assert next proposal is an eval, which we reject.

        Args:
            args: expected evaluation args (if not None)
            pred: test predicate to run on proposal (if not None)
            skip: if True, skip over all None proposals
        """
        logger.debug("Script: reject eval")
        proposal = self.proposal(skip=skip)
        self.check_eval(proposal, args=args, pred=pred).reject()

    def reject_kill(self, r=None, skip=False):
        """Assert next proposal is a kill, which we reject.

        Args:
            r: record to be killed.
            skip: if True, skip over all None proposals
        """
        logger.debug("Script: reject kill")
        self.check_kill(self.proposal(skip=skip), r).reject()

    def reject_terminate(self, skip=False):
        """Assert next proposal is a terminate, which we reject.

        Args:
            skip: if True, skip over all None proposals
        """
        logger.debug("Script: reject terminate")
        self.check_terminate(self.proposal(skip=skip)).reject()

    def terminate(self):
        "Terminate the script."
        logger.debug("Script: Terminate")
        self.call_term_callbacks()


class Monitor(object):
    """Monitor events observed by a controller.

    The monitor object provides hooks to monitor the progress of an
    optimization run by a controller.  Users should inherit from Monitor
    and add custom version of the methods

        on_new_feval(self, record)
        on_update(self, record)
        on_complete(self, record)
        on_kill(self, record)
        on_cancel(self, record)
        on_terminate(self)
    """

    def __init__(self, controller):
        """Initialize the monitor.

        Args:
            controller: The controller whose fevals we will monitor
        """
        self.controller = controller
        controller.add_feval_callback(self._add_on_update)
        controller.add_feval_callback(self.on_new_feval)
        controller.add_term_callback(self.on_terminate)

    def _add_on_update(self, record):
        "Internal handler -- add on_update callback to all new fevals."
        record.add_callback(self.on_update)

    def on_new_feval(self, record):
        "Handle new function evaluation request."
        pass

    def on_update(self, record):
        "Handle feval update."
        if record.is_completed:
            self.on_complete(record)
        elif record.is_killed:
            self.on_kill(record)
        elif record.is_cancelled:
            self.on_cancel(record)

    def on_complete(self, record):
        "Handle feval completion"
        pass

    def on_kill(self, record):
        "Handle record killed"
        pass

    def on_cancel(self, record):
        "Handle record cancelled"
        pass

    def on_terminate(self):
        "Handle termination."
        pass



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
from numpy import complex, zeros, power

import multiprocessing
from PySide2.QtCore import QThread, Signal

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import PowerFlowResults
from GridCal.Engine.Simulations.Stochastic.monte_carlo_results import MonteCarloResults
from GridCal.Engine.Simulations.Stochastic.monte_carlo_driver import make_monte_carlo_input
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import PowerFlowOptions, single_island_pf, \
                                                                   power_flow_worker_args, power_flow_post_process
from GridCal.Engine.Core.time_series_pf_data import compile_time_circuit, BranchImpedanceMode


class LatinHypercubeSampling(QThread):
    progress_signal = Signal(float)
    progress_text = Signal(str)
    done_signal = Signal()
    name = 'Latin Hypercube'

    def __init__(self, grid: MultiCircuit, options: PowerFlowOptions, sampling_points=1000,
                 opf_time_series_results=None):
        """
        Latin Hypercube constructor
        Args:
            grid: MultiCircuit instance
            options: Power flow options
            sampling_points: number of sampling points
        """
        QThread.__init__(self)

        self.circuit = grid

        self.options = options

        self.sampling_points = sampling_points

        self.opf_time_series_results = opf_time_series_results

        self.results = None

        self.logger = Logger()

        self.pool = None

        self.returned_results = list()

        self.__cancel__ = False

    def get_steps(self):
        """
        Get time steps list of strings
        """
        p = self.results.points_number
        return ['point:' + str(l) for l in range(p)]

    def update_progress_mt(self, res):
        """
        Update multi-threaded progress
        :param res:
        :return:
        """
        t, _ = res
        progress = (t + 1) / self.sampling_points * 100
        self.progress_signal.emit(progress)
        self.returned_results.append(res)

    def run_multi_thread(self):
        """
        Run the monte carlo simulation
        @return:
        """
        # print('LHS run')
        self.__cancel__ = False

        # initialize vars
        batch_size = self.sampling_points
        n = len(self.circuit.buses)
        m = self.circuit.get_branch_number()
        n_cores = multiprocessing.cpu_count()
        self.pool = multiprocessing.Pool()

        self.progress_signal.emit(0.0)
        self.progress_text.emit('Running Latin Hypercube Sampling in parallel using ' + str(n_cores) + ' cores ...')

        lhs_results = MonteCarloResults(n, m, batch_size, name='Latin Hypercube')
        avg_res = PowerFlowResults()
        avg_res.initialize(n, m)

        # compile the multi-circuit
        numerical_circuit = self.circuit.compile_time_series()

        # perform the topological computation
        calc_inputs_dict = numerical_circuit.compute(branch_tolerance_mode=self.options.branch_impedance_tolerance_mode,
                                                     ignore_single_node_islands=self.options.ignore_single_node_islands)

        # for each partition of the profiles...
        for t_key, calc_inputs in calc_inputs_dict.items():

            # For every island, run the time series
            for island_index, numerical_island in enumerate(calc_inputs):

                lhs_results.bus_types = numerical_circuit.bus_types

                monte_carlo_input = make_monte_carlo_input(numerical_island)
                mc_time_series = monte_carlo_input(batch_size, use_latin_hypercube=True)
                Vbus = numerical_island.Vbus
                branch_rates = numerical_island.branch_rates

                # short cut the indices
                b_idx = numerical_island.original_bus_idx
                br_idx = numerical_island.original_branch_idx

                # Start jobs
                self.returned_results = list()

                t = 0
                while t < batch_size and not self.__cancel__:

                    Ysh, Ibus, Sbus = mc_time_series.get_at(t)

                    args = (t, self.options, numerical_island, Vbus, Sbus, Ibus, branch_rates)

                    self.pool.apply_async(power_flow_worker_args, (args,), callback=self.update_progress_mt)

                # wait for all jobs to complete
                self.pool.close()
                self.pool.join()

                # collect results
                self.progress_text.emit('Collecting results...')
                for t, res in self.returned_results:
                    # store circuit results at the time index 't'
                    lhs_results.S_points[t, numerical_island.original_bus_idx] = res.Sbus
                    lhs_results.V_points[t, numerical_island.original_bus_idx] = res.voltage
                    lhs_results.Sbr_points[t, numerical_island.original_branch_idx] = res.Sf
                    lhs_results.loading_points[t, numerical_island.original_branch_idx] = res.loading
                    lhs_results.losses_points[t, numerical_island.original_branch_idx] = res.losses

                # compile MC results
                self.progress_text.emit('Compiling results...')
                lhs_results.compile()

                # compute the island branch results
                island_avg_res = numerical_island.compute_branch_results(lhs_results.voltage[b_idx])

                # apply the island averaged results
                avg_res.apply_from_island(island_avg_res, b_idx=b_idx, br_idx=br_idx)

        # lhs_results the averaged branch magnitudes
        lhs_results.sbranch = avg_res.Sf
        lhs_results.losses = avg_res.losses
        self.results = lhs_results

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

        return lhs_results

    def run_single_thread(self):
        """
        Run the monte carlo simulation
        @return:
        """
        # print('LHS run')
        self.__cancel__ = False

        # initialize the grid time series results
        # we will append the island results with another function

        # batch_size = self.sampling_points

        self.progress_signal.emit(0.0)
        self.progress_text.emit('Running Latin Hypercube Sampling...')

        # compile the multi-circuit
        numerical_circuit = compile_time_circuit(circuit=self.circuit,
                                                 apply_temperature=False,
                                                 branch_tolerance_mode=BranchImpedanceMode.Specified,
                                                 opf_results=self.opf_time_series_results)

        # do the topological computation
        calculation_inputs = numerical_circuit.split_into_islands(ignore_single_node_islands=self.options.ignore_single_node_islands)

        lhs_results = MonteCarloResults(n=numerical_circuit.nbus,
                                        m=numerical_circuit.nbr,
                                        p=self.sampling_points,
                                        bus_names=numerical_circuit.bus_names,
                                        branch_names=numerical_circuit.branch_names,
                                        bus_types=numerical_circuit.bus_types,
                                        name='Latin Hypercube')

        avg_res = PowerFlowResults(n=numerical_circuit.nbus,
                                   m=numerical_circuit.nbr,
                                   n_tr=numerical_circuit.ntr,
                                   n_hvdc=numerical_circuit.nhvdc,
                                   bus_names=numerical_circuit.bus_names,
                                   branch_names=numerical_circuit.branch_names,
                                   transformer_names=numerical_circuit.tr_names,
                                   hvdc_names=numerical_circuit.hvdc_names,
                                   bus_types=numerical_circuit.bus_types)

        it = 0

        # For every island, run the time series
        for island_index, numerical_island in enumerate(calculation_inputs):

            # try:
            # set the time series as sampled in the circuit
            # build the inputs
            monte_carlo_input = make_monte_carlo_input(numerical_island)
            mc_time_series = monte_carlo_input(self.sampling_points, use_latin_hypercube=True)
            Vbus = numerical_island.Vbus[:, 0]

            # short cut the indices
            bus_idx = numerical_island.original_bus_idx
            br_idx = numerical_island.original_branch_idx

            # run the time series
            for t in range(self.sampling_points):

                # set the power values from a Monte carlo point at 't'
                Y, I, S = mc_time_series.get_at(t)

                # Run the set monte carlo point at 't'
                res = single_island_pf(circuit=numerical_island,
                                       Vbus=Vbus,
                                       Sbus=S,
                                       Ibus=I,
                                       branch_rates=numerical_island.branch_rates,
                                       options=self.options,
                                       logger=self.logger)

                # Gather the results
                lhs_results.S_points[t, bus_idx] = S
                lhs_results.V_points[t, bus_idx] = res.voltage
                lhs_results.Sbr_points[t, br_idx] = res.Sf
                lhs_results.loading_points[t, br_idx] = res.loading
                lhs_results.losses_points[t, br_idx] = res.losses

                it += 1
                self.progress_signal.emit(it / self.sampling_points * 100)

                if self.__cancel__:
                    break

            if self.__cancel__:
                break

            # compile MC results
            self.progress_text.emit('Compiling results...')
            lhs_results.compile()

            # compute the island branch results
            Sfb, Stb, If, It, Vbranch, loading, \
            losses, flow_direction, Sbus = power_flow_post_process(numerical_island,
                                                                   Sbus=lhs_results.S_points.mean(axis=0)[bus_idx],
                                                                   V=lhs_results.V_points.mean(axis=0)[bus_idx],
                                                                   branch_rates=numerical_island.branch_rates)

            # apply the island averaged results
            avg_res.Sbus[bus_idx] = Sbus
            avg_res.voltage[bus_idx] = lhs_results.voltage[bus_idx]
            avg_res.Sf[br_idx] = Sfb
            avg_res.St[br_idx] = Stb
            avg_res.If[br_idx] = If
            avg_res.It[br_idx] = It
            avg_res.Vbranch[br_idx] = Vbranch
            avg_res.loading[br_idx] = loading
            avg_res.losses[br_idx] = losses
            avg_res.flow_direction[br_idx] = flow_direction

        self.results = lhs_results

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

        return lhs_results

    def run(self):
        """
        Run the monte carlo simulation
        @return:
        """
        # print('LHS run')
        self.__cancel__ = False

        if self.options.multi_thread:
            self.results = self.run_multi_thread()
        else:
            self.results = self.run_single_thread()

        # send the finnish signal
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Done!')
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the simulation
        """
        self.__cancel__ = True
        if self.pool is not None:
            self.pool.terminate()
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled')
        self.done_signal.emit()



from GridCal.Engine.Simulations.PowerFlow.power_flow_options import *
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import *
from GridCal.Engine.Simulations.PowerFlow.power_flow_driver import *
from GridCal.Engine.Simulations.PowerFlow.time_series_driver import *
from GridCal.Engine.Simulations.PowerFlow.time_Series_input import *



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui.ui',
# licensing of 'gui.ui' applies.
#
# Created: Sat May 16 17:20:30 2020
#      by: pyside2-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(933, 528)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(Dialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.tab)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.splitter = QtWidgets.QSplitter(self.tab)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.frame_4 = QtWidgets.QFrame(self.splitter)
        self.frame_4.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_4.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_4.setObjectName("frame_4")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.frame_4)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.frame = QtWidgets.QFrame(self.frame_4)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout_5.setContentsMargins(0, -1, 0, -1)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.open_button = QtWidgets.QPushButton(self.frame)
        self.open_button.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/Icons/icons/import_profiles.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.open_button.setIcon(icon)
        self.open_button.setObjectName("open_button")
        self.horizontalLayout_5.addWidget(self.open_button)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem)
        self.label_5 = QtWidgets.QLabel(self.frame)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_5.addWidget(self.label_5)
        self.units_combobox = QtWidgets.QComboBox(self.frame)
        self.units_combobox.setObjectName("units_combobox")
        self.horizontalLayout_5.addWidget(self.units_combobox)
        self.verticalLayout_5.addWidget(self.frame)
        self.sources_list = QtWidgets.QListView(self.frame_4)
        self.sources_list.setObjectName("sources_list")
        self.verticalLayout_5.addWidget(self.sources_list)
        self.frame_6 = QtWidgets.QFrame(self.splitter)
        self.frame_6.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_6.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_6.setObjectName("frame_6")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.frame_6)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, -1)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.frame_10 = QtWidgets.QFrame(self.frame_6)
        self.frame_10.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_10.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_10.setObjectName("frame_10")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout(self.frame_10)
        self.horizontalLayout_7.setContentsMargins(0, -1, 0, -1)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.autolink_button = QtWidgets.QPushButton(self.frame_10)
        self.autolink_button.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/Icons/icons/auto-link.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.autolink_button.setIcon(icon1)
        self.autolink_button.setObjectName("autolink_button")
        self.horizontalLayout_7.addWidget(self.autolink_button)
        self.autolink_slider = QtWidgets.QSlider(self.frame_10)
        self.autolink_slider.setMinimum(1)
        self.autolink_slider.setMaximum(100)
        self.autolink_slider.setProperty("value", 60)
        self.autolink_slider.setOrientation(QtCore.Qt.Horizontal)
        self.autolink_slider.setObjectName("autolink_slider")
        self.horizontalLayout_7.addWidget(self.autolink_slider)
        self.rnd_link_pushButton = QtWidgets.QPushButton(self.frame_10)
        self.rnd_link_pushButton.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/Icons/icons/auto-link-rnd.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.rnd_link_pushButton.setIcon(icon2)
        self.rnd_link_pushButton.setObjectName("rnd_link_pushButton")
        self.horizontalLayout_7.addWidget(self.rnd_link_pushButton)
        self.assign_to_all_pushButton = QtWidgets.QPushButton(self.frame_10)
        self.assign_to_all_pushButton.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/Icons/icons/link-to-all.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.assign_to_all_pushButton.setIcon(icon3)
        self.assign_to_all_pushButton.setObjectName("assign_to_all_pushButton")
        self.horizontalLayout_7.addWidget(self.assign_to_all_pushButton)
        self.assign_to_selection_pushButton = QtWidgets.QPushButton(self.frame_10)
        self.assign_to_selection_pushButton.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/Icons/icons/link-to-selection.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.assign_to_selection_pushButton.setIcon(icon4)
        self.assign_to_selection_pushButton.setObjectName("assign_to_selection_pushButton")
        self.horizontalLayout_7.addWidget(self.assign_to_selection_pushButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem1)
        self.clear_selection_button = QtWidgets.QPushButton(self.frame_10)
        self.clear_selection_button.setText("")
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/Icons/icons/delete.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.clear_selection_button.setIcon(icon5)
        self.clear_selection_button.setObjectName("clear_selection_button")
        self.horizontalLayout_7.addWidget(self.clear_selection_button)
        self.verticalLayout_3.addWidget(self.frame_10)
        self.assignation_table = QtWidgets.QTableView(self.frame_6)
        self.assignation_table.setObjectName("assignation_table")
        self.verticalLayout_3.addWidget(self.assignation_table)
        self.frame_9 = QtWidgets.QFrame(self.frame_6)
        self.frame_9.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_9.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_9.setObjectName("frame_9")
        self.gridLayout = QtWidgets.QGridLayout(self.frame_9)
        self.gridLayout.setContentsMargins(0, -1, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem2, 0, 3, 1, 1)
        self.set_multiplier_button = QtWidgets.QPushButton(self.frame_9)
        self.set_multiplier_button.setText("")
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/Icons/icons/multiply.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.set_multiplier_button.setIcon(icon6)
        self.set_multiplier_button.setObjectName("set_multiplier_button")
        self.gridLayout.addWidget(self.set_multiplier_button, 0, 0, 1, 1)
        self.multSpinBox = QtWidgets.QDoubleSpinBox(self.frame_9)
        self.multSpinBox.setMinimum(-99999.0)
        self.multSpinBox.setMaximum(99999.0)
        self.multSpinBox.setObjectName("multSpinBox")
        self.gridLayout.addWidget(self.multSpinBox, 0, 1, 1, 1)
        self.doit_button = QtWidgets.QPushButton(self.frame_9)
        self.doit_button.setStatusTip("")
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(":/Icons/icons/gear.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.doit_button.setIcon(icon7)
        self.doit_button.setObjectName("doit_button")
        self.gridLayout.addWidget(self.doit_button, 0, 5, 1, 1)
        self.normalized_checkBox = QtWidgets.QCheckBox(self.frame_9)
        self.normalized_checkBox.setObjectName("normalized_checkBox")
        self.gridLayout.addWidget(self.normalized_checkBox, 0, 4, 1, 1)
        self.verticalLayout_3.addWidget(self.frame_9)
        self.horizontalLayout_2.addWidget(self.splitter)
        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.tab_2)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.splitter_3 = QtWidgets.QSplitter(self.tab_2)
        self.splitter_3.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_3.setObjectName("splitter_3")
        self.frame_8 = QtWidgets.QFrame(self.splitter_3)
        self.frame_8.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_8.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_8.setObjectName("frame_8")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.frame_8)
        self.verticalLayout_7.setContentsMargins(-1, 0, -1, -1)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.tableView = QtWidgets.QTableView(self.frame_8)
        self.tableView.setObjectName("tableView")
        self.verticalLayout_7.addWidget(self.tableView)
        self.PlotFrame = QtWidgets.QFrame(self.splitter_3)
        self.PlotFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.PlotFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.PlotFrame.setObjectName("PlotFrame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.PlotFrame)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.plotwidget = MatplotlibWidget(self.PlotFrame)
        self.plotwidget.setObjectName("plotwidget")
        self.horizontalLayout.addWidget(self.plotwidget)
        self.horizontalLayout_4.addWidget(self.splitter_3)
        self.tabWidget.addTab(self.tab_2, "")
        self.verticalLayout.addWidget(self.tabWidget)

        self.retranslateUi(Dialog)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QtWidgets.QApplication.translate("Dialog", "Dialog", None, -1))
        self.open_button.setToolTip(QtWidgets.QApplication.translate("Dialog", "Import file", None, -1))
        self.label_5.setText(QtWidgets.QApplication.translate("Dialog", "Units", None, -1))
        self.autolink_button.setToolTip(QtWidgets.QApplication.translate("Dialog", "Automatic link profiles to objects based on their name", None, -1))
        self.autolink_slider.setToolTip(QtWidgets.QApplication.translate("Dialog", "Auto-link string simmilarity", None, -1))
        self.rnd_link_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Random-link profiles and objects", None, -1))
        self.assign_to_all_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Assign profile to all objects", None, -1))
        self.assign_to_selection_pushButton.setToolTip(QtWidgets.QApplication.translate("Dialog", "Assign profile to object selection", None, -1))
        self.clear_selection_button.setToolTip(QtWidgets.QApplication.translate("Dialog", "Clear selection", None, -1))
        self.set_multiplier_button.setToolTip(QtWidgets.QApplication.translate("Dialog", "Set multiplier", None, -1))
        self.doit_button.setToolTip(QtWidgets.QApplication.translate("Dialog", "Do it!", None, -1))
        self.doit_button.setText(QtWidgets.QApplication.translate("Dialog", "Accept", None, -1))
        self.normalized_checkBox.setToolTip(QtWidgets.QApplication.translate("Dialog", "Check if you want the profiles to be normalized on the base object property", None, -1))
        self.normalized_checkBox.setText(QtWidgets.QApplication.translate("Dialog", "normalized", None, -1))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QtWidgets.QApplication.translate("Dialog", "Assignation", None, -1))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QtWidgets.QApplication.translate("Dialog", "Plot", None, -1))

from .matplotlibwidget import MatplotlibWidget
from .icons_rc import *

if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())




# Copyright (c) 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE_MATPOWER file.

# Copyright 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# Copyright (c) 2016-2017 by University of Kassel and Fraunhofer Institute for Wind Energy and
# Energy System Technology (IWES), Kassel. All rights reserved. Use of this source code is governed
# by a BSD-style license that can be found in the LICENSE file.

# The file has been modified from Pypower.
# The function mu() has been added to the solver in order to provide an optimal iteration control
#
# Copyright (c) 2018 Santiago Peñate Vera
#
# This file retains the BSD-Style license

import time
import scipy
import scipy.sparse as sp
import numpy as np

from GridCal.Engine.Sparse.csc import pack_4_by_4
from GridCal.Engine.Simulations.sparse_solve import get_sparse_type, get_linear_solver
from GridCal.Engine.Simulations.PowerFlow.numba_functions import calc_power_csr_numba, diag
from GridCal.Engine.Simulations.PowerFlow.high_speed_jacobian import _create_J_with_numba, get_fastest_jacobian_function
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults

linear_solver = get_linear_solver()
sparse = get_sparse_type()
scipy.ALLOW_THREADS = True
np.set_printoptions(precision=8, suppress=True, linewidth=320)


def dSbus_dV(Ybus, V, I):
    """
    Computes partial derivatives of power injection w.r.t. voltage.
    :param Ybus: Admittance matrix
    :param V: Bus voltages array
    :param I: Bus current injections array
    :return:
    """
    '''
    Computes partial derivatives of power injection w.r.t. voltage.

    Returns two matrices containing partial derivatives of the complex bus
    power injections w.r.t voltage magnitude and voltage angle respectively
    (for all buses). If C{Ybus} is a sparse matrix, the return values will be
    also. The following explains the expressions used to form the matrices::

        Ibus = Ybus * V - I

        S = diag(V) * conj(Ibus) = diag(conj(Ibus)) * V

    Partials of V & Ibus w.r.t. voltage magnitudes::
        dV/dVm = diag(V / abs(V))
        dI/dVm = Ybus * dV/dVm = Ybus * diag(V / abs(V))

    Partials of V & Ibus w.r.t. voltage angles::
        dV/dVa = j * diag(V)
        dI/dVa = Ybus * dV/dVa = Ybus * j * diag(V)

    Partials of S w.r.t. voltage magnitudes::
        dS/dVm = diag(V) * conj(dI/dVm) + diag(conj(Ibus)) * dV/dVm
               = diag(V) * conj(Ybus * diag(V / abs(V)))
                                        + conj(diag(Ibus)) * diag(V / abs(V))

    Partials of S w.r.t. voltage angles::
        dS/dVa = diag(V) * conj(dI/dVa) + diag(conj(Ibus)) * dV/dVa
               = diag(V) * conj(Ybus * j * diag(V))
                                        + conj(diag(Ibus)) * j * diag(V)
               = -j * diag(V) * conj(Ybus * diag(V))
                                        + conj(diag(Ibus)) * j * diag(V)
               = j * diag(V) * conj(diag(Ibus) - Ybus * diag(V))

    For more details on the derivations behind the derivative code used
    in PYPOWER information, see:

    [TN2]  R. D. Zimmerman, "AC Power Flows, Generalized OPF Costs and
    their Derivatives using Complex Matrix Notation", MATPOWER
    Technical Note 2, February 2010.
    U{http://www.pserc.cornell.edu/matpower/TN2-OPF-Derivatives.pdf}

    @author: Ray Zimmerman (PSERC Cornell)
    '''

    ib = range(len(V))

    Ibus = Ybus * V - I
    diagV = sparse((V, (ib, ib)))
    diagIbus = sparse((Ibus, (ib, ib)))
    diagVnorm = sparse((V / np.abs(V), (ib, ib)))

    dS_dVm = diagV * np.conj(Ybus * diagVnorm) + np.conj(diagIbus) * diagVnorm
    dS_dVa = 1.0j * diagV * np.conj(diagIbus - Ybus * diagV)

    return dS_dVm, dS_dVa


def mu(Ybus, Ibus, J, pvpq_lookup, incS, dV, dx, pvpq, pq, npv, npq):
    """
    Calculate the Iwamoto acceleration parameter as described in:
    "A Load Flow Calculation Method for Ill-Conditioned Power Systems" by Iwamoto, S. and Tamura, Y."
    Args:
        Ybus: Admittance matrix
        J: Jacobian matrix
        incS: mismatch vector
        dV: voltage increment (in complex form)
        dx: solution vector as calculated dx = solve(J, incS)
        pvpq: array of the pq and pv indices
        pq: array of the pq indices

    Returns:
        the Iwamoto's optimal multiplier for ill conditioned systems
    """
    # evaluate the Jacobian of the voltage derivative
    # theoretically this is the second derivative matrix
    # since the Jacobian (J2) has been calculated with dV instead of V

    # generate lookup pvpq -> index pvpq (used in createJ)
    # pvpq_lookup = np.zeros(np.max(Ybus.indices) + 1, dtype=int)
    # pvpq_lookup[pvpq] = np.arange(len(pvpq))

    J2 = _create_J_with_numba(Ybus, dV, pvpq, pq, pvpq_lookup, npv, npq)
    # J2 = Jacobian(Ybus, dV, Ibus, pq, pvpq)

    a = incS
    b = J * dx
    c = 0.5 * dx * J2 * dx

    g0 = -a.dot(b)
    g1 = b.dot(b) + 2 * a.dot(c)
    g2 = -3.0 * b.dot(c)
    g3 = 2.0 * c.dot(c)

    roots = np.roots([g3, g2, g1, g0])
    # three solutions are provided, the first two are complex, only the real solution is valid
    return roots[2].real


def Jacobian(Ybus, V, Ibus, pq, pvpq):
    """
    Computes the system Jacobian matrix in polar coordinates
    Args:
        Ybus: Admittance matrix
        V: Array of nodal voltages
        Ibus: Array of nodal current injections
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses

    Returns:
        The system Jacobian matrix
    """
    I = Ybus * V - Ibus

    diagV = sp.diags(V)
    diagI = sp.diags(I)
    diagVnorm = sp.diags(V / np.abs(V))

    dS_dVm = diagV * np.conj(Ybus * diagVnorm) + np.conj(diagI) * diagVnorm
    dS_dVa = 1.0j * diagV * np.conj(diagI - Ybus * diagV)

    J = sp.vstack([sp.hstack([dS_dVa[np.ix_(pvpq, pvpq)].real, dS_dVm[np.ix_(pvpq, pq)].real]),
                   sp.hstack([dS_dVa[np.ix_(pq, pvpq)].imag,   dS_dVm[np.ix_(pq, pq)].imag])], format="csc")

    return sparse(J)


def Jacobian_cartesian(Ybus, V, Ibus, pq, pvpq):
    """
    Computes the system Jacobian matrix in cartesian coordinates
    Args:
        Ybus: Admittance matrix
        V: Array of nodal voltages
        Ibus: Array of nodal current injections
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses

    Returns:
        The system Jacobian matrix in cartesian coordinates
    """
    I = Ybus * V - Ibus

    diagV = sp.diags(V)
    diagI = sp.diags(I)
    VY = diagV * np.conj(Ybus)

    dS_dVr = np.conj(diagI) + VY  # dSbus / dVr
    dS_dVi = 1j * (np.conj(diagI) - VY)  # dSbus / dVi

    '''
    j11 = real(dSbus_dVr([pq; pv], pq));    j12 = real(dSbus_dVi([pq; pv], [pv; pq]));
    
    j21 = imag(dSbus_dVr(pq, pq));          j22 = imag(dSbus_dVi(pq, [pv; pq]));
    

    J = [   j11 j12;
            j21 j22;    ];
    '''

    J = sp.vstack([sp.hstack([dS_dVr[np.ix_(pvpq, pq)].real, dS_dVi[np.ix_(pvpq, pvpq)].real]),
                   sp.hstack([dS_dVr[np.ix_(pq, pq)].imag,   dS_dVi[np.ix_(pq, pvpq)].imag])], format="csc")

    return sparse(J)


def Jacobian_decoupled(Ybus, V, Ibus, pq, pvpq):
    """
    Computes the decouped Jacobian matrices
    Args:
        Ybus: Admittance matrix
        V: Array of nodal voltages
        Ibus: Array of nodal current injections
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses

    Returns: J11, J22
    """
    I = Ybus * V - Ibus

    diagV = sp.diags(V)
    diagI = sp.diags(I)
    diagVnorm = sp.diags(V / np.abs(V))

    dS_dVm = diagV * np.conj(Ybus * diagVnorm) + np.conj(diagI) * diagVnorm
    dS_dVa = 1.0j * diagV * np.conj(diagI - Ybus * diagV)

    J11 = dS_dVa[np.ix_(pvpq, pvpq)].real
    J22 = dS_dVm[np.ix_(pq, pq)].imag

    return J11, J22


def calc_power(V, Ybus, Ibus):
    """

    :param V:
    :param Ybus:
    :param Ibus:
    :return:
    """
    # S = V * np.conj(Ybus * V - Ibus)
    # Y = Ybus.tocsr()
    S = calc_power_csr_numba(n=V.shape[0], Yp=Ybus.indptr, Yj=Ybus.indices, Yx=Ybus.data, V=V, I=Ibus, n_par=500)

    return S


def NR_LS(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15, mu_0=1.0,
          acceleration_parameter=0.05, error_registry=None) -> NumericPowerFlowResults:
    """
    Solves the power flow using a full Newton's method with backtrack correction.
    @Author: Santiago Peñate Vera
    :param Ybus: Admittance matrix
    :param Sbus: Array of nodal power injections
    :param V0: Array of nodal voltages (initial solution)
    :param Ibus: Array of nodal current injections
    :param pv: Array with the indices of the PV buses
    :param pq: Array with the indices of the PQ buses
    :param tol: Tolerance
    :param max_it: Maximum number of iterations
    :param mu_0: initial acceleration value
    :param acceleration_parameter: parameter used to correct the "bad" iterations, should be be between 1e-3 ~ 0.5
    :param error_registry: list to store the error for plotting
    :return: Voltage solution, converged?, error, calculated power injections
    """
    start = time.time()

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv and pq buses
    j1 = 0
    j2 = npv + npq
    # j2:j3 - V mag of pq buses
    j3 = j2 + npq

    if (npq + npv) > 0:

        # generate lookup pvpq -> index pvpq (used in createJ)
        pvpq_lookup = np.zeros(np.max(Ybus.indices) + 1, dtype=int)
        pvpq_lookup[pvpq] = np.arange(len(pvpq))

        # evaluate F(x0)
        Scalc = V * np.conj(Ybus * V - Ibus)
        dS = Scalc - Sbus  # compute the mismatch
        f = np.r_[dS[pvpq].real, dS[pq].imag]

        # check tolerance
        norm_f = 0.5 * f.dot(f)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f < tol:
            converged = 1

        # to be able to compare
        Ybus.sort_indices()

        # do Newton iterations
        while not converged and iter_ < max_it:
            # update iteration counter
            iter_ += 1

            # evaluate Jacobian
            # J = Jacobian(Ybus, V, Ibus, pq, pvpq)
            J = _create_J_with_numba(Ybus, V, pvpq, pq, pvpq_lookup, npv, npq)

            # compute update step
            dx = linear_solver(J, f)

            # reassign the solution vector
            dVa[pvpq] = dx[j1:j2]
            dVm[pq] = dx[j2:j3]

            # update voltage the Newton way (mu=1)
            mu_ = mu_0
            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            Vnew = Vm * np.exp(1.0j * Va)

            # compute the mismatch function f(x_new)
            Scalc = Vnew * np.conj(Ybus * Vnew - Ibus)
            dS = Scalc - Sbus  # complex power mismatch
            f_new = np.r_[dS[pvpq].real, dS[pq].imag]  # concatenate to form the mismatch function
            norm_f_new = 0.5 * f_new.dot(f_new)

            if error_registry is not None:
                error_registry.append(norm_f_new)

            cond = norm_f_new > norm_f  # condition to back track (no improvement at all)

            if not cond:
                back_track_counter += 1

            l_iter = 0
            while cond and l_iter < max_it and mu_ > tol:

                # reset to the previous voltage
                Va = np.angle(V)
                Vm = np.abs(V)

                # line search back
                # update voltage with a closer value to the last value in the Jacobian direction
                mu_ *= acceleration_parameter
                Vm -= mu_ * dVm
                Va -= mu_ * dVa
                Vnew = Vm * np.exp(1.0j * Va)

                # compute the mismatch function f(x_new)
                Scalc = Vnew * np.conj(Ybus * Vnew - Ibus)
                dS = Scalc - Sbus  # complex power mismatch
                f_new = np.r_[dS[pvpq].real, dS[pq].imag]  # concatenate to form the mismatch function

                norm_f_new = 0.5 * f_new.dot(f_new)

                cond = norm_f_new > norm_f

                if error_registry is not None:
                    error_registry.append(norm_f_new)

                l_iter += 1
                back_track_iterations += 1

            # update calculation variables
            V = Vnew
            f = f_new

            # check for convergence
            if l_iter == 0:
                # no correction loop executed, hence compute the error fresh
                norm_f = 0.5 * f_new.dot(f_new)
            else:
                # pick the latest computer error in the correction loop
                norm_f = norm_f_new

            if error_registry is not None:
                error_registry.append(norm_f)

            if norm_f < tol:
                converged = 1

    else:
        norm_f = 0
        converged = True
        Scalc = Sbus

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, norm_f, Scalc, None, None, None, iter_, elapsed)


def NRD_LS(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15,
           acceleration_parameter=0.5, error_registry=None) -> NumericPowerFlowResults:
    """
    Solves the power flow using a full Newton's method with backtrack correction.
    @Author: Santiago Peñate Vera
    :param Ybus: Admittance matrix
    :param Sbus: Array of nodal power injections
    :param V0: Array of nodal voltages (initial solution)
    :param Ibus: Array of nodal current injections
    :param pv: Array with the indices of the PV buses
    :param pq: Array with the indices of the PQ buses
    :param tol: Tolerance
    :param max_it: Maximum number of iterations
    :param acceleration_parameter: parameter used to correct the "bad" iterations, typically 0.5
    :param error_registry: list to store the error for plotting
    :return: Voltage solution, converged?, error, calculated power injections
    """

    start = time.time()

    use_norm_error = True

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # evaluate F(x0)
    Scalc = V * np.conj(Ybus * V - Ibus)
    dS = Scalc - Sbus  # compute the mismatch
    f = np.r_[dS[pvpq].real, dS[pq].imag]

    # check tolerance
    if use_norm_error:
        norm_f = np.linalg.norm(f, np.Inf)
    else:
        norm_f = 0.5 * f.dot(f)

    if error_registry is not None:
        error_registry.append(norm_f)

    if norm_f < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J1, J4 = Jacobian_decoupled(Ybus, V, Ibus, pq, pvpq)

        # compute update step and reassign the solution vector
        dVa[pvpq] = linear_solver(J1, f[pvpq])
        dVm[pq] = linear_solver(J4, f[pq])

        # update voltage the Newton way (mu=1)
        mu_ = 1.0
        Vm -= mu_ * dVm
        Va -= mu_ * dVa
        Vnew = Vm * np.exp(1.0j * Va)

        # compute the mismatch function f(x_new)
        dS = Vnew * np.conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
        f_new = np.r_[dS[pvpq].real, dS[pq].imag]  # concatenate to form the mismatch function

        if use_norm_error:
            norm_f_new = np.linalg.norm(f_new, np.Inf)
        else:
            norm_f_new = 0.5 * f_new.dot(f_new)

        if error_registry is not None:
            error_registry.append(norm_f_new)

        cond = norm_f_new > norm_f  # condition to back track (no improvement at all)

        if not cond:
            back_track_counter += 1

        l_iter = 0
        while not cond and l_iter < 10 and mu_ > 0.01:
            # line search back
            # update voltage with a closer value to the last value in the Jacobian direction
            mu_ *= acceleration_parameter
            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            Vnew = Vm * np.exp(1.0j * Va)

            # compute the mismatch function f(x_new)
            dS = Vnew * np.conj(Ybus * Vnew - Ibus) - Sbus  # complex power mismatch
            f_new = np.r_[dS[pvpq].real, dS[pq].imag]  # concatenate to form the mismatch function

            if use_norm_error:
                norm_f_new = np.linalg.norm(f_new, np.Inf)
            else:
                norm_f_new = 0.5 * f_new.dot(f_new)

            cond = norm_f_new > norm_f

            if error_registry is not None:
                error_registry.append(norm_f_new)

            l_iter += 1
            back_track_iterations += 1

        # update calculation variables
        V = Vnew
        f = f_new

        # check for convergence
        if use_norm_error:
            norm_f = np.linalg.norm(f_new, np.Inf)
        else:
            norm_f = 0.5 * f_new.dot(f_new)

        if error_registry is not None:
            error_registry.append(norm_f)

        if norm_f < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    # print('iter_', iter_, '  -  back_track_counter', back_track_counter,
    #       '  -  back_track_iterations', back_track_iterations)

    return NumericPowerFlowResults(V, converged, norm_f, Scalc, None, None, None, iter_, elapsed)


def IwamotoNR(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15, robust=False) -> NumericPowerFlowResults:
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        robust: Boolean variable for the use of the Iwamoto optimal step factor.
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @author: Ray Zimmerman (PSERC Cornell)
    @Author: Santiago Penate Vera
    """
    start = time.time()

    # initialize
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv + npq
    # j2:j3 - V mag of pq buses
    j3 = j2 + npq

    if (npq + npv) > 0:

        # generate lookup pvpq -> index pvpq (used in createJ)
        pvpq_lookup = np.zeros(np.max(Ybus.indices) + 1, dtype=int)
        pvpq_lookup[pvpq] = np.arange(len(pvpq))

        # evaluate F(x0)
        Scalc = V * np.conj(Ybus * V - Ibus)
        mis = Scalc - Sbus  # compute the mismatch
        f = np.r_[mis[pvpq].real, mis[pq].imag]

        # check tolerance
        norm_f = np.linalg.norm(f, np.Inf)

        if norm_f < tol:
            converged = 1

        # do Newton iterations
        while not converged and iter_ < max_it:
            # update iteration counter
            iter_ += 1

            # evaluate Jacobian
            # J = Jacobian(Ybus, V, Ibus, pq, pvpq)
            # Ybus, V, pvpq, pq, pvpq_lookup, npv, npq
            J = _create_J_with_numba(Ybus, V, pvpq, pq, pvpq_lookup, npv, npq)

            # compute update step
            try:
                dx = linear_solver(J, f)
            except:
                print(J)
                converged = False
                iter_ = max_it + 1  # exit condition

            # reassign the solution vector
            dVa[pvpq] = dx[j1:j2]
            dVm[pq] = dx[j2:j3]
            dV = dVm * np.exp(1j * dVa)  # voltage mismatch

            # update voltage
            if robust:
                # if dV contains zeros will crash the second Jacobian derivative
                if not (dV == 0.0).any():
                    # calculate the optimal multiplier for enhanced convergence
                    mu_ = mu(Ybus, Ibus, J, pvpq_lookup, f, dV, dx, pvpq, pq, npv, npq)
                else:
                    mu_ = 1.0
            else:
                mu_ = 1.0

            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            V = Vm * np.exp(1j * Va)

            Vm = np.abs(V)  # update Vm and Va again in case
            Va = np.angle(V)  # we wrapped around with a negative Vm

            # evaluate F(x)
            Scalc = V * np.conj(Ybus * V - Ibus)
            mis = Scalc - Sbus  # complex power mismatch
            f = np.r_[mis[pvpq].real, mis[pq].imag]  # concatenate again

            # check for convergence
            norm_f = np.linalg.norm(f, np.Inf)
            converged = norm_f < tol
    else:
        norm_f = 0
        converged = True
        Scalc = Sbus

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, norm_f, Scalc, None, None, None, iter_, elapsed)


def levenberg_marquardt_pf(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=50) -> NumericPowerFlowResults:
    """
    Solves the power flow problem by the Levenberg-Marquardt power flow algorithm.
    It is usually better than Newton-Raphson, but it takes an order of magnitude more time to converge.
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Peñate Vera
    """
    start = time.time()

    # initialize
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv and pq buses
    j1 = 0
    j2 = npv + npq
    # j2:j3 - V mag of pq buses
    j3 = j2 + npq

    if (npq + npv) > 0:
        normF = 100000
        update_jacobian = True
        converged = False
        iter_ = 0
        nu = 2.0
        lbmda = 0
        f_prev = 1e9  # very large number
        nn = 2 * npq + npv
        ii = np.linspace(0, nn-1, nn)
        Idn = sparse((np.ones(nn), (ii, ii)), shape=(nn, nn))  # csc_matrix identity

        # generate lookup pvpq -> index pvpq (used in createJ)
        pvpq_lookup = np.zeros(np.max(Ybus.indices) + 1, dtype=int)
        pvpq_lookup[pvpq] = np.arange(len(pvpq))

        while not converged and iter_ < max_it:

            # evaluate Jacobian
            if update_jacobian:
                H = _create_J_with_numba(Ybus, V, pvpq, pq, pvpq_lookup, npv, npq)
                # H = Jacobian(Ybus, V, Ibus, pq, pvpq)

            # evaluate the solution error F(x0)
            Scalc = V * np.conj(Ybus * V - Ibus)
            mis = Scalc - Sbus  # compute the mismatch
            dz = np.r_[mis[pvpq].real, mis[pq].imag]  # mismatch in the Jacobian order

            # system matrix
            # H1 = H^t
            H1 = H.transpose().tocsr()

            # H2 = H1·H
            H2 = H1.dot(H)

            # set first value of lmbda
            if iter_ == 0:
                lbmda = 1e-3 * H2.diagonal().max()

            # compute system matrix A = H^T·H - lambda·I
            A = H2 + lbmda * Idn

            # right hand side
            # H^t·dz
            rhs = H1.dot(dz)

            # Solve the increment
            dx = linear_solver(A, rhs)

            # objective function to minimize
            f = 0.5 * dz.dot(dz)

            # decision function
            val = dx.dot(lbmda * dx + rhs)
            if val > 0.0:
                rho = (f_prev - f) / (0.5 * val)
            else:
                rho = -1.0

            # lambda update
            if rho >= 0:
                update_jacobian = True
                lbmda *= max([1.0 / 3.0, 1 - (2 * rho - 1) ** 3])
                nu = 2.0

                # reassign the solution vector
                dVa[pvpq] = dx[j1:j2]
                dVm[pq] = dx[j2:j3]

                # update Vm and Va again in case we wrapped around with a negative Vm
                Vm -= dVm
                Va -= dVa
                V = Vm * np.exp(1.0j * Va)

            else:
                update_jacobian = False
                lbmda *= nu
                nu *= 2.0

            # check convergence
            Scalc = V * np.conj(Ybus.dot(V))
            ds = Sbus - Scalc
            e = np.r_[ds[pvpq].real, ds[pq].imag]
            normF = 0.5 * np.dot(e, e)
            converged = normF < tol
            f_prev = f

            # update iteration counter
            iter_ += 1
    else:
        normF = 0
        converged = True
        Scalc = Sbus  # V * np.conj(Ybus * V - Ibus)
        iter_ = 0

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, normF, Scalc, None, None, None, iter_, elapsed)


def Jacobian_I(Ybus, V, pq, pvpq):
    """
    Computes the system Jacobian matrix
    Args:
        Ybus: Admittance matrix
        V: Array of nodal voltages
        pq: Array with the indices of the PQ buses
        pvpq: Array with the indices of the PV and PQ buses

    Returns:
        The system Jacobian matrix in current equations
    """
    dI_dVm = Ybus * sp.diags(V / np.abs(V))
    dI_dVa = 1j * (Ybus * sp.diags(V))

    J11 = dI_dVa[np.array([pvpq]).T, pvpq].real
    J12 = dI_dVm[np.array([pvpq]).T, pq].real
    J21 = dI_dVa[np.array([pq]).T, pvpq].imag
    J22 = dI_dVm[np.array([pq]).T, pq].imag

    J = sp.vstack([sp.hstack([J11, J12]),
                   sp.hstack([J21, J22])], format="csr")

    return J


def condition_number(J):
    """
    Computes the condition number of a sparse matrix
    :param J: sparse matrix like the Jacobian
    :return:
    """
    norm_A = scipy.sparse.linalg.norm(J)
    norm_invA = scipy.sparse.linalg.norm(scipy.sparse.linalg.inv(J))
    cond = norm_A * norm_invA
    return cond


def NR_I_LS(Ybus, Sbus_sp, V0, Ibus_sp, pv, pq, tol, max_it=15, acceleration_parameter=0.5) -> NumericPowerFlowResults:
    """
    Solves the power flow using a full Newton's method in current equations with current mismatch with line search
    Args:
        Ybus: Admittance matrix
        Sbus_sp: Array of nodal specified power injections
        V0: Array of nodal voltages (initial solution)
        Ibus_sp: Array of nodal specified current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        acceleration_parameter: value used to correct bad iterations
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """
    start = time.time()

    # initialize
    back_track_counter = 0
    back_track_iterations = 0
    alpha = 1e-4
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Icalc = Ybus * V - Ibus_sp
    dI = np.conj(Sbus_sp / V) - Icalc  # compute the mismatch
    F = np.r_[dI[pvpq].real, dI[pq].imag]
    normF = np.linalg.norm(F, np.Inf)  # check tolerance

    if normF < tol:
        converged = 1

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        # evaluate Jacobian
        J = Jacobian_I(Ybus, V, pq, pvpq)

        # compute update step
        dx = linear_solver(J, F)

        # reassign the solution vector
        dVa[pvpq] = dx[j1:j4]
        dVm[pq] = dx[j5:j6]

        # update voltage the Newton way (mu=1)
        mu_ = 1.0
        Vm += mu_ * dVm
        Va += mu_ * dVa
        Vnew = Vm * np.exp(1j * Va)

        # compute the mismatch function f(x_new)
        Icalc = Ybus * Vnew - Ibus_sp
        dI = np.conj(Sbus_sp / Vnew) - Icalc
        Fnew = np.r_[dI[pvpq].real, dI[pq].imag]

        normFnew = np.linalg.norm(Fnew, np.Inf)

        cond = normF < normFnew  # condition to back track (no improvement at all)

        if not cond:
            back_track_counter += 1

        l_iter = 0
        while cond and l_iter < max_it and mu_ > tol:
            # line search back

            # reset voltage
            Va = np.angle(V)
            Vm = np.abs(V)

            # update voltage with a closer value to the last value in the Jacobian direction
            mu_ *= acceleration_parameter
            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            Vnew = Vm * np.exp(1j * Va)

            # compute the mismatch function f(x_new)
            Icalc = Ybus * Vnew - Ibus_sp
            dI = np.conj(Sbus_sp / Vnew) - Icalc
            Fnew = np.r_[dI[pvpq].real, dI[pq].imag]

            normFnew = np.linalg.norm(Fnew, np.Inf)
            cond = normF < normFnew

            l_iter += 1
            back_track_iterations += 1

        # update calculation variables
        V = Vnew
        F = Fnew

        # check for convergence
        normF = normFnew

        if normF < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    Scalc = V * np.conj(Icalc)

    return NumericPowerFlowResults(V, converged, normF, Scalc, None, None, None, iter_, elapsed)


def F(V, Ybus, S, I, pq, pvpq):
    """

    :param V:
    :param Ybus:
    :param S:
    :param I:
    :param pq:
    :param pv:
    :param pvpq:
    :return:
    """
    # compute the mismatch function f(x_new)
    dS = V * np.conj(Ybus * V - I) - S  # complex power mismatch
    return np.r_[dS[pvpq].real, dS[pq].imag]  # concatenate to form the mismatch function


def fx(x, Ybus, S, I, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm):
    """

    :param x:
    :param Ybus:
    :param S:
    :param I:
    :param pq:
    :param pv:
    :param pvpq:
    :return:
    """
    n = len(S)
    npv = len(pv)
    npq = len(pq)

    # reassign the solution vector
    if npv:
        Va[pv] = x[j1:j2]
    if npq:
        Va[pq] = x[j3:j4]
        Vm[pq] = x[j5:j6]

    V = Vm * np.exp(1j * Va)  # voltage mismatch

    # right hand side
    g = F(V, Ybus, S, I, pq, pvpq)

    # jacobian
    gx = Jacobian(Ybus, V, I, pq, pvpq)

    # return the increment of x
    return linear_solver(gx, g)


def ContinuousNR(Ybus, Sbus, V0, Ibus, pv, pq, tol, max_it=15) -> NumericPowerFlowResults:
    """
    Solves the power flow using a full Newton's method with the backtrack improvement algorithm
    Args:
        Ybus: Admittance matrix
        Sbus: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        Ibus: Array of nodal current injections
        pv: Array with the indices of the PV buses
        pq: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        robust: Boolean variable for the use of the Iwamoto optimal step factor.
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @author: Ray Zimmerman (PSERC Cornell)
    @Author: Santiago Penate Vera
    """

    start = time.time()

    # initialize
    converged = 0
    iter_ = 0
    V = V0.copy()

    # set up indexing for updating V
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)

    # j1:j2 - V angle of pv buses
    j1 = 0
    j2 = npv
    # j3:j4 - V angle of pq buses
    j3 = j2
    j4 = j2 + npq
    # j5:j6 - V mag of pq buses
    j5 = j4
    j6 = j4 + npq

    # evaluate F(x0)
    Scalc = V * np.conj(Ybus * V - Ibus)
    mis = Scalc - Sbus  # compute the mismatch
    F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]

    # check tolerance
    normF = np.linalg.norm(F, np.Inf)

    if normF < tol:
        converged = 1

    dt = 1.0

    # Compose x
    x = np.zeros(2 * npq + npv)
    Va = np.angle(V)
    Vm = np.abs(V)

    # do Newton iterations
    while not converged and iter_ < max_it:
        # update iteration counter
        iter_ += 1

        x[j1:j4] = Va[pvpq]
        x[j5:j6] = Vm[pq]

        # Compute the Runge-Kutta steps
        k1 = fx(x,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        k2 = fx(x + 0.5 * dt * k1,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        k3 = fx(x + 0.5 * dt * k2,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        k4 = fx(x + dt * k3,
                Ybus, Sbus, Ibus, pq, pv, pvpq, j1, j2, j3, j4, j5, j6, Va, Vm)

        x -= dt * (k1 + 2.0 * k2 + 2.0 * k3 + k4) / 6.0

        # reassign the solution vector
        Va[pvpq] = x[j1:j4]
        Vm[pq] = x[j5:j6]
        V = Vm * np.exp(1j * Va)  # voltage mismatch

        # evaluate F(x)
        Scalc = V * np.conj(Ybus * V - Ibus)
        mis = Scalc - Sbus  # complex power mismatch
        F = np.r_[mis[pv].real, mis[pq].real, mis[pq].imag]  # concatenate again

        # check for convergence
        normF = np.linalg.norm(F, np.Inf)

        if normF > 0.01:
            dt = max(dt * 0.985, 0.75)
        else:
            dt = min(dt * 1.015, 0.75)

        print(dt)

        if normF < tol:
            converged = 1

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, normF, Scalc, None, None, None, iter_, elapsed)




import networkx as nx
from matplotlib import pyplot as plt

from research.three_phase.Engine.circuit_matrix_inputs import *


class Circuit:

    def __init__(self, Sbase=100, n_phase=3):
        """
        Circuit constructor
        This circuit must be fully connected
        :param Sbase: Base power in MVA
        :param n_phase: Number of phases of the circuit
        """

        # circuit base power in MVA
        self.Sbase = Sbase

        # number of phases of the circuit
        self.n_phase = n_phase

        # list of buses
        self.buses = list()

        # list of branches (lines, transformers, etc.)
        self.branches = list()

        # dictionary of bus objects to their index [obj] -> index
        self.bus_idx = dict()

        # circuit Graph
        self.graph = None

    @staticmethod
    def apply_ABCD(k, f, t, phases_from, phases_to, n_phase, Ybus, Yseries, Cf, Ct,
                   mask_r, mask_c, mask_k,
                   A, B, C, D, As, Ds):
        """
        Generic function to apply the A, B, C, D sub-matrices to the larger Ybus matrix
        :param k: index of the branch in the circuit
        :param f: index of the "from" bus in the circuit
        :param t: index of the "to" bus in the circuit
        :param phases_from: vector of connection phases of the "from" bus
        :param phases_to: vector of connection phases of the "to" bus
        :param n_phase: number of phases of the circuit
        :param Ybus: Circuit admittance matrix (sparse and complex)
        :param Yf: Circuit branch- "from" bus admittance matrix
        :param Yt: Circuit branch- "to" bus admittance matrix
        :param Cf: Circuit branch- "from" bus connectivity matrix
        :param Ct: Circuit branch- "to" bus connectivity matrix
        :param mask_r: mask for the rows of Ybus
        :param mask_c: mask for the columns of Ybus
        :param mask_k: mask for the branches
        :param A: admittance sub-matrix "A"
        :param B: admittance sub-matrix "B"
        :param C: admittance sub-matrix "C"
        :param D: admittance sub-matrix "D"
        :return: Nothing, The matrices Ybus, Cf, Ct, and the masks are modified in-place (bu reference)
        """
        # Set the A, B, C, D matrices into the larger Ybus Matrix
        a = 0
        for i in phases_from:
            b = 0
            for j in phases_to:

                # Admittance matrix
                Ybus[f * n_phase + i, f * n_phase + j] += A[a, b]
                Ybus[f * n_phase + i, t * n_phase + j] += B[a, b]
                Ybus[t * n_phase + i, f * n_phase + j] += C[a, b]
                Ybus[t * n_phase + i, t * n_phase + j] += D[a, b]

                Yseries[f * n_phase + i, f * n_phase + j] += As[a, b]
                Yseries[f * n_phase + i, t * n_phase + j] += B[a, b]
                Yseries[t * n_phase + i, f * n_phase + j] += C[a, b]
                Yseries[t * n_phase + i, t * n_phase + j] += Ds[a, b]

                # branch-bus admittance matrices
                # Yf[k * n_phase + i, f * n_phase + j] += A[a, b]
                # Yf[k * n_phase + i, t * n_phase + j] += B[a, b]
                # Yt[k * n_phase + i, f * n_phase + j] += C[a, b]
                # Yt[k * n_phase + i, t * n_phase + j] += D[a, b]

                # branch-bus connectivity matrices
                Cf[k * n_phase + i, f * n_phase + j] = 1
                Cf[k * n_phase + i, t * n_phase + j] = -1
                Ct[k * n_phase + i, f * n_phase + j] = -1
                Ct[k * n_phase + i, t * n_phase + j] = 1

                # masks
                mask_r[f * n_phase + i] = 1
                mask_r[t * n_phase + i] = 1
                mask_c[f * n_phase + j] = 1
                mask_c[t * n_phase + j] = 1
                mask_k[k * n_phase + i] = 1

                b += 1
            a += 1

    def compile(self):
        """
        Compile the circuit into matrices
        :return: circuit data object (instance of CircuitMatrices)
        """
        n = len(self.buses)
        m = len(self.branches)
        self.graph = nx.Graph()

        # declare the matrices container object
        data = CircuitMatrixInputs(n, m, self.n_phase)

        # compile nodes
        for k, elm in enumerate(self.buses):

            # collect the node magnitudes in per unit
            elm.apply_YISV(k, data, self.Sbase, n_phase=self.n_phase)

            # store the object and its index in a dictionary
            self.bus_idx[elm] = k

            # set the node entry in a graph
            self.graph.add_node(elm.name, pos=(elm.x, elm.y))

        # Add the bus shunt admittances vector into Ybus diagonal
        ib = range(len(data.Vbus))
        data.Ybus[ib, ib] += data.Ybus_sh

        # compile branches
        for k, elm in enumerate(self.branches):

            # get the from and to bus indices from the connection objects of the branch element
            f = self.bus_idx[elm.f]
            t = self.bus_idx[elm.t]

            self.graph.add_edge(elm.f.name, elm.t.name)

            # apply the ABCD parameters in per unit
            A, B, C, D, As, Ds = elm.get_ABCD(self.Sbase)
            self.apply_ABCD(k, f, t,
                            phases_from=elm.phases_from,
                            phases_to=elm.phases_to,
                            n_phase=self.n_phase,
                            Ybus=data.Ybus,
                            Yseries=data.Yseries,
                            Cf=data.Cf,
                            Ct=data.Ct,
                            mask_r=data.mask_r,
                            mask_c=data.mask_c,
                            mask_k=data.mask_k,
                            A=A, B=B, C=C, D=D, As=As, Ds=Ds)

            # get the rating
            for i in range(self.n_phase):
                data.branch_rates[k * self.n_phase + i] = elm.rating

        # consolidate the data: Reduce the matrices, find the bus type lists, etc...
        data.consolidate()

        return data

    def pack_node_solution(self, vec):
        """
        Pack a nodal vector to match the buses
        :param vec: nodal vector (can be non symmetric)
        :return: packed vector by bus
        """
        n = len(self.buses)
        val = [None] * n

        a = 0
        for k, elm in enumerate(self.buses):
            val[k] = vec[a:a + elm.number_of_phases]
            a += elm.number_of_phases

        return val

    def pack_branch_solution(self, vec):
        """
        Pack a nodal vector to match the buses
        :param vec: nodal vector (can be non symmetric)
        :return: packed vector by bus
        """
        m = len(self.branches)
        val = [None] * m

        a = 0
        for k, elm in enumerate(self.branches):
            val[k] = vec[a:a + elm.number_of_phases]
            a += elm.number_of_phases

        return val

    def plot(self):
        """
        Plot the circuit layout
        :return:
        """
        # from networkx.drawing.nx_pylab import draw_networkx
        # fig = plt.Figure(figsize=(16, 12))
        # ax = fig.add_subplot(111)
        pos = nx.get_node_attributes(self.graph, 'pos')
        nx.draw(self.graph, pos,  with_labels=True, figsize=(16, 12))
        plt.show()




from GridCal.Engine.Core.topology import Graph

# from GridCal.Engine.Core.snapshot_pf_data import SnapshotData
# from GridCal.Engine.Core.time_series_pf_data import TimeCircuit
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.admittance_matrices import Admittance




# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.
import os
import h5py
from h5py._hl.dataset import Dataset
from h5py._hl.group import Group
import pandas as pd
import numpy as np

from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.IO.excel_interface import create_data_frames, interpret_excel_v3


def save_h5(circuit: MultiCircuit, file_path):
    """
    Save the circuit information in excel format
    :param circuit: MultiCircuit instance
    :param file_path: path to the excel file
    :return: logger with information
    """
    logger = Logger()

    dfs = create_data_frames(circuit=circuit)

    store = pd.HDFStore(file_path)

    for key in dfs.keys():
        store[key] = dfs[key]

    return logger


def open_h5(file_path):

    circuit = MultiCircuit()

    store = pd.HDFStore(file_path)

    dfs = dict()
    for group in store.root:
        dfs[group._v_name] = pd.read_hdf(store, group._v_pathname)

    return dfs


def save_dict_to_hdf5(dic, filename):
    """

    :param dic:
    :param filename:
    :return:
    """
    with h5py.File(filename, 'w') as h5file:
        recursively_save_dict_contents_to_group(h5file, '/', dic)


def recursively_save_dict_contents_to_group(h5file, path, dic):
    """

    :param h5file:
    :param path:
    :param dic:
    :return:
    """
    for key, item in dic.items():
        if isinstance(item, (np.ndarray, np.int64, np.float64, str, bytes)):
            h5file[path + key] = item
        elif isinstance(item, dict):
            recursively_save_dict_contents_to_group(h5file, path + key + '/', item)
        else:
            raise ValueError('Cannot save %s type'%type(item))


def load_dict_from_hdf5(filename):
    """

    :param filename:
    :return:
    """
    with h5py.File(filename, 'r') as h5file:
        return recursively_load_dict_contents_from_group(h5file, '/')


def recursively_load_dict_contents_from_group(h5file, path):
    """

    :param h5file:
    :param path:
    :return:
    """
    ans = {}
    for key, item in h5file[path].items():
        if isinstance(item, Dataset):
            ans[key] = item.value
        elif isinstance(item, Group):
            ans[key] = recursively_load_dict_contents_from_group(h5file, path + key + '/')
    return ans


if __name__ == '__main__':

    from GridCal.Engine.IO.file_handler import *

    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'

    circuit = FileOpen(fname).open()

    save_h5(circuit, file_path='1354 pegase.h5')

    circuit2 = open_h5('1354 pegase.h5')



"""
This is the linear ac power flow from the article:

Linearized AC Load Flow Applied to Analysis in Electric Power Systems
by: P. Rossoni, W. M da Rosa and E. A. Belati

Implementation by Santiago Peñate Vera 2018
"""

import time
import numpy as np
np.set_printoptions(linewidth=320)
from numpy import zeros, ones, mod, conj, array, r_, linalg, Inf, complex128, c_, r_, angle
from itertools import product
from numpy.linalg import solve
from scipy.sparse.linalg import spsolve
from scipy.sparse import issparse, csc_matrix as sparse
from scipy.sparse import hstack as hstack_s, vstack as vstack_s
import pandas as pd


def LACPF(Y, Ys, S, Vset, pq, pv):
    """
    Linearized AC Load Flow
    Args:
        Y: Admittance matrix
        Ys: Admittance matrix of the series elements
        S: Power injections vector of all the nodes
        Vset: Set voltages of all the nodes (used for the slack and PV nodes)
        pq: list of indices of the pq nodes
        pv: list of indices of the pv nodes

    Returns: Voltage vector and error
    """

    pvpq = r_[pv, pq]
    npq = len(pq)
    npv = len(pv)

    # compose the system matrix
    G = Y.real
    B = Y.imag
    Gp = Ys.real
    Bp = Ys.imag

    A11 = -Bp[np.ix_(pvpq, pvpq)]
    A12 = G[np.ix_(pvpq, pq)]
    A21 = -Gp[np.ix_(pq, pvpq)]
    A22 = -B[np.ix_(pq, pq)]

    Asys = vstack_s([hstack_s([A11, A12]),
                     hstack_s([A21, A22])], format="csc")

    # compose the right hand side (power vectors)
    rhs = r_[S.real[pvpq], S.imag[pq]]

    # solve the linear system
    x = spsolve(Asys, rhs)

    # compose the results vector
    voltages_vector = Vset.copy()

    #  set the pv voltages
    va_pv = x[0:npv]
    vm_pv = np.abs(Vset[pv])
    voltages_vector[pv] = vm_pv * np.exp(1.0j * va_pv)

    # set the PQ voltages
    va_pq = x[npv:npv+npq]
    vm_pq = np.ones(npq) - x[npv+npq::]
    voltages_vector[pq] = vm_pq * np.exp(1.0j * va_pq)

    # Calculate the error and check the convergence
    Scalc = voltages_vector * conj(Y * voltages_vector)

    # complex power mismatch
    power_mismatch = Scalc - S

    # concatenate error by type
    mismatch = r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]

    # check for convergence
    normF = linalg.norm(mismatch, Inf)

    return voltages_vector, normF


def LACPF_2(Y, Ys, S, Vset, pq, pv):
    """
    Linearized AC Load Flow
    Args:
        Y: Admittance matrix
        Ys: Admittance matrix of the series elements
        S: Power injections vector of all the nodes
        Vset: Set voltages of all the nodes (used for the slack and PV nodes)
        pq: list of indices of the pq nodes
        pv: list of indices of the pv nodes

    Returns: Voltage vector and error
    """

    pvpq = r_[pv, pq]
    npq = len(pq)
    npv = len(pv)

    # compose the system matrix
    G = Y.real
    B = Y.imag
    Gp = Ys.real
    Bp = Ys.imag

    A11 = -Bp[np.ix_(pvpq, pvpq)]
    A12 = G[np.ix_(pvpq, pq)]
    A21 = -Gp[np.ix_(pq, pvpq)]
    A22 = -B[np.ix_(pq, pq)]

    Asys = vstack_s([hstack_s([A11, A12]),
                     hstack_s([A21, A22])], format="csc")

    # compose the right hand side (power vectors)
    rhs = r_[S.real[pvpq], S.imag[pq]]

    # solve the linear system
    x = spsolve(Asys, rhs)

    # compose the results vector
    voltages_vector = Vset.copy()

    #  set the pv voltages
    va_pv = x[0:npv]
    vm_pv = np.abs(Vset[pv])
    voltages_vector[pv] = vm_pv * np.exp(1.0j * va_pv)

    # set the PQ voltages
    va_pq = x[npv:npv + npq]
    vm_pq = np.ones(npq) - x[npv + npq::]
    voltages_vector[pq] = vm_pq * np.exp(1.0j * va_pq)

    # Calculate the error and check the convergence
    Scalc = voltages_vector * conj(Y * voltages_vector)

    # complex power mismatch
    power_mismatch = Scalc - S

    # ------------------------------------------------------------------------------------------------------------------
    drhs = r_[power_mismatch.real[pvpq], power_mismatch.imag[pq]]
    dx = spsolve(Asys, drhs)

    #  set the pv voltages
    va_pv = x[0:npv] + dx[0:npv]
    vm_pv = np.abs(Vset[pv])
    voltages_vector[pv] = vm_pv * np.exp(1.0j * va_pv)

    # set the PQ voltages
    va_pq = x[npv:npv + npq] + dx[npv:npv + npq]
    vm_pq = np.ones(npq) - x[npv + npq::] - dx[npv + npq::]
    voltages_vector[pq] = vm_pq * np.exp(1.0j * va_pq)

    # ------------------------------------------------------------------------------------------------------------------

    # concatenate error by type
    mismatch = r_[power_mismatch[pv].real, power_mismatch[pq].real, power_mismatch[pq].imag]

    # check for convergence
    normF = linalg.norm(mismatch, Inf)

    return voltages_vector, normF


def res_2_df(V, Sbus, tpe):
    """
    Create dataframe to display the results nicely
    :param V: Voltage complex vector
    :param Sbus: Power complex vector
    :param tpe: Types
    :return: Pandas DataFrame
    """
    vm = abs(V)
    va = np.angle(V)

    d = {1: 'PQ', 2: 'PV', 3: 'VD'}

    tpe_str = array([d[i] for i in tpe], dtype=object)
    data = c_[tpe_str, Sbus.real, Sbus.imag, vm, va]
    cols = ['Type', 'P', 'Q', '|V|', 'angle']
    df = pd.DataFrame(data=data, columns=cols)

    return df


if __name__ == '__main__':
    from GridCal.Engine import FileOpen, PowerFlowOptions, PowerFlowDriver, SolverType
    from matplotlib import pyplot as plt


    # grid.load_file('lynn5buspq.xlsx')
    # grid.load_file('lynn5buspv.xlsx')
    # grid.load_file('IEEE30.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39.xlsx')
    # grid.load_file('/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx')

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE 14.xlsx'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/1354 Pegase.xlsx'


    grid = FileOpen(fname).open()

    numerical_circuit = grid.compile_snapshot()

    circuits = numerical_circuit.compute()

    circuit = circuits[0]

    print('\nYbus:\n', circuit.Ybus.todense())
    print('\nSbus:\n', circuit.Sbus)
    print('\nIbus:\n', circuit.Ibus)
    print('\nVbus:\n', circuit.Vbus)
    print('\ntypes:\n', circuit.types)
    print('\npq:\n', circuit.pq)
    print('\npv:\n', circuit.pv)
    print('\nvd:\n', circuit.ref)

    start_time = time.time()

    # Y, Ys, Ysh, max_coefficient_count, S, voltage_set_points, pq, pv, vd
    v, err = LACPF_2(Y=circuit.Ybus,
                     Ys=circuit.Yseries,
                     S=circuit.Sbus,
                     Vset=circuit.Vbus,
                     pq=circuit.pq,
                     pv=circuit.pv)

    print('Linear AC:')
    print("--- %s seconds ---" % (time.time() - start_time))
    print('Results:\n', res_2_df(v, circuit.Sbus, circuit.types))
    print('error: \t', err)

    # check the method solution: v against the Newton-Raphson power flow
    print('\nNR')
    options = PowerFlowOptions(SolverType.NR, verbose=False, tolerance=1e-9, control_q=False)
    power_flow = PowerFlowDriver(grid, options)

    start_time = time.time()
    power_flow.run()
    print("--- %s seconds ---" % (time.time() - start_time))
    vnr = power_flow.results.voltage

    print('Results:\n', res_2_df(vnr, circuit.Sbus, circuit.types))
    print('error: \t', power_flow.results.error)

    # check
    dif = v - vnr
    print('\ndiff:\t', dif)
    print('\nmax diff:\t', max(abs(dif)))

    plt.show()



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class MeasurementType(Enum):
    Pinj = "Active power injection",
    Qinj = "Reactive power injection",
    Vmag = "Voltage magnitude",
    Pflow = "Active power flow",
    Qflow = "Reactive power flow",
    Iflow = "Current module flow"


class Measurement:

    def __init__(self, value, uncertainty, mtype: MeasurementType, idtag=None):
        """
        Constructor
        :param value: value
        :param uncertainty: uncertainty (standard deviation)
        :param mtype: type of measurement
        """
        self.val = value
        self.sigma = uncertainty
        self.measurement_type = mtype
        idtag = idtag



"""
This example is coming from the book:
Power System Load Flow Analysis - Lynn Powell

The added profile is just to demonstrate how to create load profiles properly

Author: Santiago Peñate Vera (September 2018)
"""
import numpy as np
import pandas as pd
from GridCal.Engine import *
from GridCal.Engine.Devices.branch import Branch, BranchTemplate
from GridCal.Engine.Devices.bus import Bus
from GridCal.Engine.Devices.generator import Generator
from GridCal.Engine.Devices.load import Load
from GridCal.Engine.Devices.enumerations import BranchType
from GridCal.Engine.Simulations.PowerFlow.power_flow_worker import PowerFlowOptions, SolverType
from GridCal.Engine.Simulations.PowerFlow.power_flow_driver import PowerFlowDriver
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Simulations.PowerFlow.time_series_driver import TimeSeries


def main():
    ####################################################################################################################
    # Define the circuit
    #
    # A circuit contains all the grid information regardless of the islands formed or the amount of devices
    ####################################################################################################################

    # create a circuit

    grid = MultiCircuit(name='lynn 5 bus')

    # let's create a master profile

    st = datetime.datetime(2020, 1, 1)
    dates = [st + datetime.timedelta(hours=i) for i in range(24)]
    time_array = pd.to_datetime(dates)
    x = np.linspace(-np.pi, np.pi, len(time_array))
    y = np.abs(np.sin(x))
    df_0 = pd.DataFrame(data=y, index=time_array)  # complex values

    # set the grid master time profile
    grid.time_profile = df_0.index

    ####################################################################################################################
    # Define the buses
    ####################################################################################################################
    # I will define this bus with all the properties so you see
    bus1 = Bus(name='Bus1',
               vnom=10,   # Nominal voltage in kV
               vmin=0.9,  # Bus minimum voltage in per unit
               vmax=1.1,  # Bus maximum voltage in per unit
               xpos=0,    # Bus x position in pixels
               ypos=0,    # Bus y position in pixels
               height=0,  # Bus height in pixels
               width=0,   # Bus width in pixels
               active=True,   # Is the bus active?
               is_slack=False,  # Is this bus a slack bus?
               area='Defualt',  # Area (for grouping purposes only)
               zone='Default',  # Zone (for grouping purposes only)
               substation='Default'  # Substation (for grouping purposes only)
               )

    # the rest of the buses are defined with the default parameters
    bus2 = Bus(name='Bus2')
    bus3 = Bus(name='Bus3')
    bus4 = Bus(name='Bus4')
    bus5 = Bus(name='Bus5')

    # add the bus objects to the circuit
    grid.add_bus(bus1)
    grid.add_bus(bus2)
    grid.add_bus(bus3)
    grid.add_bus(bus4)
    grid.add_bus(bus5)

    ####################################################################################################################
    # Add the loads
    ####################################################################################################################

    # In GridCal, the loads, generators ect are stored within each bus object:

    # we'll define the first load completely
    l2 = Load(name='Load',
              G=0, B=0,  # admittance of the ZIP model in MVA at the nominal voltage
              Ir=0, Ii=0,  # Current of the ZIP model in MVA at the nominal voltage
              P=40, Q=20,  # Power of the ZIP model in MVA
              active=True,  # Is active?
              mttf=0.0,  # Mean time to failure
              mttr=0.0  # Mean time to recovery
              )
    grid.add_load(bus2, l2)

    # Define the others with the default parameters
    grid.add_load(bus3, Load(P=25, Q=15))
    grid.add_load(bus4, Load(P=40, Q=20))
    grid.add_load(bus5, Load(P=50, Q=20))

    ####################################################################################################################
    # Add the generators
    ####################################################################################################################

    g1 = Generator(name='gen',
                   active_power=0.0,  # Active power in MW, since this generator is used to set the slack , is 0
                   voltage_module=1.0,  # Voltage set point to control
                   Qmin=-9999,  # minimum reactive power in MVAr
                   Qmax=9999,  # Maximum reactive power in MVAr
                   Snom=9999,  # Nominal power in MVA
                   power_prof=None,  # power profile
                   vset_prof=None,  # voltage set point profile
                   active=True  # Is active?
                   )
    grid.add_generator(bus1, g1)

    ####################################################################################################################
    # Add the lines
    ####################################################################################################################

    br1 = Branch(bus_from=bus1,
                 bus_to=bus2,
                 name='Line 1-2',
                 r=0.05,  # resistance of the pi model in per unit
                 x=0.11,  # reactance of the pi model in per unit
                 g=1e-20,  # conductance of the pi model in per unit
                 b=0.02,  # susceptance of the pi model in per unit
                 rate=50,  # Rate in MVA
                 tap=1.0,  # Tap value (value close to 1)
                 shift_angle=0,  # Tap angle in radians
                 active=True,  # is the branch active?
                 mttf=0,  # Mean time to failure
                 mttr=0,  # Mean time to recovery
                 branch_type=BranchType.Line,  # Branch type tag
                 length=1,  # Length in km (to be used with templates)
                 template=BranchTemplate()  # Branch template (The default one is void)
                 )
    grid.add_branch(br1)

    grid.add_branch(Branch(bus1, bus3, name='Line 1-3', r=0.05, x=0.11, b=0.02, rate=50))
    grid.add_branch(Branch(bus1, bus5, name='Line 1-5', r=0.03, x=0.08, b=0.02, rate=80))
    grid.add_branch(Branch(bus2, bus3, name='Line 2-3', r=0.04, x=0.09, b=0.02, rate=3))
    grid.add_branch(Branch(bus2, bus5, name='Line 2-5', r=0.04, x=0.09, b=0.02, rate=10))
    grid.add_branch(Branch(bus3, bus4, name='Line 3-4', r=0.06, x=0.13, b=0.03, rate=30))
    grid.add_branch(Branch(bus4, bus5, name='Line 4-5', r=0.04, x=0.09, b=0.02, rate=30))

    FileSave(grid, 'lynn5node.gridcal').save()

    ####################################################################################################################
    # Overwrite the default profiles with the custom ones
    ####################################################################################################################

    for load in grid.get_loads():
        load.P_prof = load.P * df_0.values[:, 0]
        load.Q_prof = load.Q * df_0.values[:, 0]

    for gen in grid.get_static_generators():
        gen.P_prof = gen.Q * df_0.values[:, 0]
        gen.Q_prof = gen.Q * df_0.values[:, 0]

    for gen in grid.get_generators():
        gen.P_prof = gen.P * df_0.values[:, 0]

    ####################################################################################################################
    # Run a power flow simulation
    ####################################################################################################################

    # We need to specify power flow options
    pf_options = PowerFlowOptions(solver_type=SolverType.NR,  # Base method to use
                                  verbose=False,  # Verbose option where available
                                  tolerance=1e-6,  # power error in p.u.
                                  max_iter=25,  # maximum iteration number
                                  control_q=True  # if to control the reactive power
                                  )

    # Declare and execute the power flow simulation
    pf = PowerFlowDriver(grid, pf_options)
    pf.run()

    writer = pd.ExcelWriter('Results.xlsx')
    # now, let's compose a nice DataFrame with the voltage results
    headers = ['Vm (p.u.)', 'Va (Deg)', 'Vre', 'Vim']
    Vm = np.abs(pf.results.voltage)
    Va = np.angle(pf.results.voltage, deg=True)
    Vre = pf.results.voltage.real
    Vim = pf.results.voltage.imag
    data = np.c_[Vm, Va, Vre, Vim]
    v_df = pd.DataFrame(data=data, columns=headers, index=grid.bus_names)
    # print('\n', v_df)
    v_df.to_excel(writer, sheet_name='V')

    # Let's do the same for the branch results
    headers = ['Loading (%)', 'Current(p.u.)', 'Power (MVA)']
    loading = np.abs(pf.results.loading) * 100
    current = np.abs(pf.results.If)
    power = np.abs(pf.results.Sf)
    data = np.c_[loading, current, power]
    br_df = pd.DataFrame(data=data, columns=headers, index=grid.branch_names)
    br_df.to_excel(writer, sheet_name='Br')

    # Finally the execution metrics
    print('\nError:', pf.results.error)
    print('Elapsed time (s):', pf.results.elapsed, '\n')

    # print(tabulate(v_df, tablefmt="pipe", headers=v_df.columns.values))
    # print()
    # print(tabulate(br_df, tablefmt="pipe", headers=br_df.columns.values))



    ####################################################################################################################
    # Run a time series power flow simulation
    ####################################################################################################################

    ts = TimeSeries(grid=grid,
                    options=pf_options,
                    opf_time_series_results=None,
                    start_=0,
                    end_=None)

    ts.run()

    print()
    print('-' * 200)
    print('Time series')
    print('-' * 200)
    print('Voltage time series')
    df_voltage = pd.DataFrame(data=np.abs(ts.results.voltage), columns=grid.bus_names, index=grid.time_profile)
    df_voltage.to_excel(writer, sheet_name='Vts')
    writer.close()


if __name__ == '__main__':
    main()



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

from GridCal.Engine.Simulations.PTDF.ptdf_driver import *
from GridCal.Engine.Simulations.PTDF.ptdf_analysis import *
from GridCal.Engine.Simulations.PTDF.ptdf_results import *
from GridCal.Engine.Simulations.PTDF.ptdf_ts_driver import *
from GridCal.Engine.Simulations.PTDF.analytic_ptdf import *
from GridCal.Engine.Simulations.PTDF.analytic_ptdf_driver import *



# This file is part of GridCal.
#
# GridCal is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GridCal is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GridCal.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from GridCal.Engine.Devices import DeviceType


class ResultTypes(Enum):
    # Power flow
    BusVoltage = 'Bus voltage', DeviceType.BusDevice
    BusVoltagePolar = 'Bus voltage (polar)', DeviceType.BusDevice
    BusActivePower = 'Bus active power', DeviceType.BusDevice
    BusReactivePower = 'Bus reactive power', DeviceType.BusDevice
    BranchPower = 'Branch power', DeviceType.BranchDevice
    BranchActivePowerFrom = 'Branch active power "from"', DeviceType.BranchDevice
    BranchReactivePowerFrom = 'Branch reactive power "from"', DeviceType.BranchDevice
    BranchActivePowerTo = 'Branch active power "to"', DeviceType.BranchDevice
    BranchReactivePowerTo = 'Branch reactive power "to"', DeviceType.BranchDevice

    BranchCurrent = 'Branch current', DeviceType.BranchDevice
    BranchActiveCurrentFrom = 'Branch active current "from"', DeviceType.BranchDevice
    BranchReactiveCurrentFrom = 'Branch reactive current "from"', DeviceType.BranchDevice
    BranchActiveCurrentTo = 'Branch active current "to"', DeviceType.BranchDevice
    BranchReactiveCurrentTo = 'Branch reactive current "to"', DeviceType.BranchDevice

    BranchTapModule = 'Branch tap module', DeviceType.BranchDevice
    BranchTapAngle = 'Branch tap angle', DeviceType.BranchDevice
    BranchBeq = 'Branch Beq', DeviceType.BranchDevice

    BranchLoading = 'Branch loading', DeviceType.BranchDevice
    Transformer2WTapModule = 'Transformer tap module', DeviceType.Transformer2WDevice
    BranchVoltage = 'Branch voltage drop', DeviceType.BranchDevice
    BranchAngles = 'Branch voltage angles', DeviceType.BranchDevice
    BranchLosses = 'Branch losses', DeviceType.BranchDevice
    BranchActiveLosses = 'Branch active losses', DeviceType.BranchDevice
    BranchReactiveLosses = 'Branch reactive losses', DeviceType.BranchDevice
    BatteryPower = 'Battery power', DeviceType.BatteryDevice
    BatteryEnergy = 'Battery energy', DeviceType.BatteryDevice
    HvdcLosses = 'HVDC losses', DeviceType.HVDCLineDevice
    HvdcSentPower = 'HVDC sent power', DeviceType.HVDCLineDevice

    # MonteCarlo
    BusVoltageAverage = 'Bus voltage avg', DeviceType.BusDevice
    BusVoltageStd = 'Bus voltage std', DeviceType.BusDevice
    BusVoltageCDF = 'Bus voltage CDF', DeviceType.BusDevice
    BusPowerCDF = 'Bus power CDF', DeviceType.BusDevice
    BranchPowerAverage = 'Branch power avg', DeviceType.BranchDevice
    BranchPowerStd = 'Branch power std', DeviceType.BranchDevice
    BranchPowerCDF = 'Branch power CDF', DeviceType.BranchDevice
    BranchLoadingAverage = 'Branch loading avg', DeviceType.BranchDevice
    BranchLoadingStd = 'Branch loading std', DeviceType.BranchDevice
    BranchLoadingCDF = 'Branch loading CDF', DeviceType.BranchDevice
    BranchLossesAverage = 'Branch losses avg', DeviceType.BranchDevice
    BranchLossesStd = 'Branch losses std', DeviceType.BranchDevice
    BranchLossesCDF = 'Branch losses CDF', DeviceType.BranchDevice

    # OPF
    BusVoltageModule = 'Bus voltage module', DeviceType.BusDevice
    BusVoltageAngle = 'Bus voltage angle', DeviceType.BusDevice
    BusPower = 'Bus power', DeviceType.BusDevice
    ShadowPrices = 'Bus shadow prices', DeviceType.BusDevice
    BranchOverloads = 'Branch overloads', DeviceType.BranchDevice
    LoadShedding = 'Load shedding', DeviceType.LoadDevice
    ControlledGeneratorShedding = 'Generator shedding', DeviceType.GeneratorDevice
    ControlledGeneratorPower = 'Generator power', DeviceType.GeneratorDevice

    # Short-circuit
    BusShortCircuitPower = 'Bus short circuit power', DeviceType.BusDevice

    # PTDF
    PTDFBranchesSensitivity = 'Branch Flow sensitivity', DeviceType.BranchDevice
    PTDFBusVoltageSensitivity = 'Bus voltage sensitivity', DeviceType.BusDevice

    OTDF = 'Outage transfer distribution factors', DeviceType.BranchDevice

    SimulationError = 'Error', DeviceType.BusDevice

    OTDFSimulationError = 'Error', DeviceType.BranchDevice

    # sigma
    SigmaReal = 'Sigma real', DeviceType.BusDevice
    SigmaImag = 'Sigma imaginary', DeviceType.BusDevice
    SigmaDistances = 'Sigma distances', DeviceType.BusDevice
    SigmaPlusDistances = 'Sigma + distances', DeviceType.BusDevice


class SimulationTypes(Enum):
    PowerFlow_run = 'power flow'
    ShortCircuit_run = 'Short circuit'
    MonteCarlo_run = 'Monte Carlo'
    TimeSeries_run = 'Time series power flow'
    VoltageCollapse_run = 'Voltage collapse'
    LatinHypercube_run = 'Latin Hypercube'
    Cascade_run = 'Cascade'
    OPF_run = 'Optimal power flow'
    OPFTimeSeries_run = 'OPF Time series'
    TransientStability_run = 'Transient stability'
    TopologyReduction_run = 'Topology reduction'
    PTDF_run = 'PTDF'
    PTDF_TS_run = 'PTDF_TS'
    OTDF_run = 'OTDF'
    Delete_and_reduce_run = 'Delete and reduce'
